import os

from . import utils


class MergeDict(object):
    def __init__(
        self, platform=None, platforms=("linux", "windows"), site=None, **format_kwargs
    ):
        self.format_kwargs = format_kwargs
        self.formatter = self.default_format
        self.pathsep = os.pathsep
        self.platforms = platforms
        self.validator = None
        self.site = site

        if platform is None:
            platform = utils.Platform.name()
        self.platform = platform

    def apply_platform_wildcards(self, data, output=None):
        """Ensure the data dict is platform specific and any wildcard entries
        are applied to all of self.platforms.

        Args:
            data (dict): A dictionary of settings to apply. `make_os_specific`
                is called on it.
            output (dict, optional): If provided, the dictionary is updated in
                with the info stored in data. If not, a new dictionary is
                created and returned.

        Returns:
            dict: The modified dictionary. If output was provided it is the
                same object, if not a new dict is returned.
        """
        data = self.make_os_specific(data)

        if output is None:
            output = {"os_specific": True}

        for platform in self.platforms:
            platform_data = output.setdefault(platform, {})

            if "*" in data:
                self.update_platform(platform_data, data["*"], platform=platform)

            if platform in data:
                self.update_platform(platform_data, data[platform], platform=platform)

        return output

    def default_format(self, value, platform=None):
        """Apply string formatting rules to the given value if applicable.

        If value is a list, this method is recursively called on the contents
        and a new list is returned. If a bool or dict are passed, they are
        returned without modification. Otherwise its assume to be a string and
        str.format is called passing `self.format_kwargs`.

        If `self.site` is set and platform is passed, site.platform_path_map is called
        on the text output to convert it to the desired platform.
        """
        if isinstance(value, list):
            # Format the individual items if a list of args is used.
            # return [v.format(**self.format_kwargs) for v in value]
            return [self.default_format(v) for v in value]
        if isinstance(value, (bool, dict, int)):
            return value
        ret = value.format(**self.format_kwargs)

        if platform and self.site:
            ret = self.site.platform_path_map(ret, platform)

        return ret

    @property
    def format_kwargs(self):
        return self._format_kwargs

    @format_kwargs.setter
    def format_kwargs(self, value):
        self._format_kwargs = value
        self._format_kwargs_cleaned = {
            k: utils.path_forward_slash(v) for k, v in value.items()
        }

    def join(self, a, b):
        """Join the two inputs into a flat list. If an input is a string it
        is split by ``self.pathsep``.

        Args:
            a: ``b`` is added after this.
            b: ``a`` is added before this.

        Returns:
            list: The joined a and b values.
        """
        if isinstance(a, str):
            a = utils.Platform.path_split(a, pathsep=self.pathsep)
        if isinstance(b, str):
            b = utils.Platform.path_split(b, pathsep=self.pathsep)

        if isinstance(a, dict):
            if isinstance(b, dict):
                a.update(b)
            return a

        return a + b

    @classmethod
    def make_os_specific(cls, data):
        """Ensure the dict conforms to the "os_specific=True" specification.
        If os_specific is missing or not true, a new dict is returned with a
        copy of data stored under the "*" key with `os_specific` removed.

        Args:
            data (dict): The dict to make os_specific. The original dict is
                returned if it's already os_specific, otherwise a new dict is
                returned.

        Returns:
            dict: A dict that conforms to the "os_specific=True" format.
        """
        if not data.get("os_specific", False):
            # Remove os_specific if it was defined so we don't keep it on the
            # wildcard platform dict.
            data = data.copy()
            data.pop("os_specific", False)
            data = {"*": data}
            data["os_specific"] = True
        return data

    def update_platform(self, data, changes, platform=None):
        if self.validator:
            self.validator(changes)

        if "unset" in changes:
            # When applying the env vars later None will trigger removing the env var.
            # The other operations may end up replacing this value.
            data.update({key: None for key in changes["unset"]})

        # set, prepend, append are all treated as set operations, this lets us override
        # base user and system variable values without them causing issues.
        if "set" in changes:
            for key, value in changes["set"].items():
                data[key] = self.formatter(value, platform=platform)
                if isinstance(data[key], str):
                    data[key] = [data[key]]

        for operation in ("prepend", "append"):
            if operation not in changes:
                continue

            for key, value in changes[operation].items():
                value = self.formatter(value, platform=platform)
                existing = data.get(key, "")
                if existing:
                    if operation == "prepend":
                        value = self.join(value, existing)
                    else:
                        value = self.join(existing, value)
                else:
                    # Convert value into a list if it isn't one
                    value = self.join(value, [])

                data[key] = value



import logging
import os
from collections import UserDict
from pathlib import Path, PurePosixPath, PureWindowsPath

from colorama import Fore, Style

from . import utils
from .cache import Cache
from .merge_dict import MergeDict

logger = logging.getLogger(__name__)


class Site(UserDict):
    """Provides site configuration to hab.

    This dictionary is updated with the contents of each json file stored in paths
    See `hab.MergeDict` for details.

    Args:
        paths (list, optional): A list of paths to json files defining how
            site should be setup. If not specified, the ``HAB_PATHS`` environment
            variable is used.
    """

    """Provides some default values that are always configured but can be updated."""
    _default_data = {
        "set": {
            "config_paths": [],
            "distro_paths": [],
            "ignored_distros": ["release", "pre"],
            "platforms": ["windows", "osx", "linux"],
            "site_cache_file_template": ["{{stem}}.habcache"],
        }
    }

    def __init__(self, paths=None, platform=None):
        if platform is None:
            platform = utils.Platform.name()
        self.platform = platform

        # Add default data to all site instances. Site data is only valid for
        # the current platform, so discard any other platform configurations.
        merger = MergeDict(platforms=[self.platform])
        self.frozen_data = merger.apply_platform_wildcards(self._default_data)

        if not paths:
            paths = os.getenv("HAB_PATHS", "").split(os.pathsep)
        self.paths = [Path(os.path.expandvars(p)).expanduser() for p in paths if p]

        self.load()

        # Create the caching class instance using the entry point
        eps = self.entry_points_for_group("hab.habcache_cls")
        if eps:
            habcache_cls = eps[0].load()
        else:
            # If not defined use the default cache class
            habcache_cls = Cache
        self.cache = habcache_cls(self)

    @property
    def data(self):
        return self.frozen_data.get(self.platform)

    def dump(self, verbosity=0, color=None, width=80):
        """Return a string of the properties and their values.

        Args:
            verbosity (int, optional): More information is shown with higher values.
            color (bool, optional): Add console colorization to output. If None,
                respect the site property "colorize" defaulting to True.
            width (int, optional): The desired width for wrapping. The output may
                exceed this value, but it will attempt to respect it.

        Returns:
            str: The configuration converted to a string
        """
        if color is None:
            color = self.get("colorize", True)

        def cached_fmt(path, cached):
            if not cached:
                return path
            if color:
                return f"{path} {Fore.YELLOW}(cached){Style.RESET_ALL}"
            else:
                return f"{path} (cached)"

        # Include the paths used to configure this site object
        hab_paths = []
        for path in self.paths:
            if verbosity:
                # Indicate if a cache file exists for each site config file.
                cache_file = self.cache.site_cache_path(path)
                path = cached_fmt(path, cache_file.is_file())
            hab_paths.append(str(path))
        site_ret = utils.dump_object({"HAB_PATHS": hab_paths}, color=color, width=width)
        # Include all of the resolved site configurations
        ret = []
        for prop, value in self.items():
            if verbosity and prop in ("config_paths", "distro_paths"):
                cache = getattr(self.cache, prop)()
                paths = []
                for dirname, _, cached in self.cache.iter_cache_paths(
                    prop, value, cache, include_path=False
                ):
                    paths.append(cached_fmt(dirname, cached))
                txt = utils.dump_object(
                    paths, label=f"{prop}:  ", color=color, width=width
                )
            elif verbosity < 1 and isinstance(value, dict):
                # This is too complex for most site dumps, hide the details behind
                # a higher verbosity setting.
                txt = utils.dump_object(
                    f"Dictionary keys: {len(value)}",
                    label=f"{prop}:  ",
                    color=color,
                    width=width,
                )
            else:
                txt = utils.dump_object(
                    value, label=f"{prop}:  ", color=color, width=width
                )

            ret.append(txt)

        ret = "\n".join(ret)
        return utils.dump_title("Dump of Site", f"{site_ret}\n{ret}", color=color)

    def entry_points_for_group(
        self, group, default=None, entry_points=None, omit_none=True
    ):
        """Returns a list of importlib_metadata.EntryPoint objects enabled by
        this site config. To import and resolve the defined object call `ep.load()`.

        Args:
            group (str): The name of the group of entry_points to process.
            default (dict, optional): If the entry_point is not defined, return
                the entry points defined by this dictionary. This is the contents
                of the entry_points group, not the entire entry_points dict. For
                example: `{"gui": "hab_gui.cli:gui"}`
            entry_points (dict, optional): Use this dictionary instead of the one
                defined on this Site object.
            omit_none (bool, optional): If an entry_point's value is set to null/None
                then don't include an EntryPoint object for it in the return. This
                allows a second site file to disable a entry_point already set.
        """
        # Delay this import to when required. It's faster than pkg_resources but
        # no need to pay the import price for it if you are not using it.
        from importlib_metadata import EntryPoint

        ret = []
        # Use the site defined entry_points if an override dict wasn't provided
        if entry_points is None:
            entry_points = self.get("entry_points", {})

        # Get the entry point definitions, falling back to default if provided
        if group in entry_points:
            ep_defs = entry_points[group]
        else:
            ep_defs = default if default else {}

        # Init the EntryPoint objects
        # While we are using importlib.metadata to create EntryPoints, we are not
        # using it's `entry_points` function. We want the current site configuration
        # to define the entry points being loaded not the installed pip packages.
        for name, value in ep_defs.items():
            if omit_none and value is None:
                continue

            ep = EntryPoint(name=name, group=group, value=value)
            ret.append(ep)
        return ret

    def load(self):
        """Iterates over each file in self.path. Replacing the value of each key.
        The last file in the list will have its settings applied even if other
        files define them."""
        # Process the main site files. These are the only ones that can add the
        # `hab.site.add_paths` entry_points.
        for path in reversed(self.paths):
            self.load_file(path)

        # Now that the main site files are handle `hab.site.add_paths` entry_points.
        # This lets you add site json files where you can't hard code the path.
        # For example if you want a site file included in a pip package installed
        # on a host, the path would change depending on the python version being
        # used and if using a editable pip install.
        for ep in self.entry_points_for_group("hab.site.add_paths"):
            logger.debug(f"Running hab.site.add_paths entry_point: {ep}")
            func = ep.load()

            # This entry_point needs to return a list of site file paths as
            # `pathlib.Path` records.
            paths = func(site=self)
            for path in reversed(paths):
                if path in self.paths:
                    logger.debug(f"Path already added, skipping: {path}")
                    continue
                logger.debug(f"Path added by hab.site.add_paths: {path}")
                self.paths.insert(0, path)
                self.load_file(path)

        # Convert config_paths and distro_paths to lists of Path objects
        self["config_paths"] = utils.Platform.expand_paths(self["config_paths"])
        self["distro_paths"] = utils.Platform.expand_paths(self["distro_paths"])

        # Ensure any platform_path_maps are converted to pathlib objects.
        self.standardize_platform_path_maps()

        # Entry_point to allow modification as a final step of loading site files
        self.run_entry_points_for_group("hab.site.finalize", site=self)

    def load_file(self, filename):
        """Load an individual file path and merge its contents onto self.

        Args:
            filename (pathlib.Path): The json file to parse.
        """
        data = utils.load_json_file(filename)

        # Merge the new data into frozen_data
        merger = MergeDict(platforms=[self.platform], relative_root=filename.parent)
        merger.apply_platform_wildcards(data, output=self.frozen_data)

    @property
    def paths(self):
        """A list of ``pathlib.Path``'s processed by load."""
        return self._paths

    @paths.setter
    def paths(self, paths):
        self._paths = paths

    def platform_path_key(self, path, platform=None):
        """Converts the provided full path to a str.format style path.

        Uses mappings defined in `site.get('platform_path_maps', {})` to convert
        full file paths to the map key name.
        """
        if self.platform == "windows":
            path = PureWindowsPath(path)
        else:
            path = PurePosixPath(path)

        platform = utils.Platform.name()
        mappings = self.get("platform_path_maps", {})

        for key in mappings:
            m = mappings[key][platform]
            try:
                relative = path.relative_to(m)
            except ValueError:
                relative = ""
            is_relative = bool(relative)
            if is_relative:
                # platform_path_maps only replace the start of a file path so
                # there is no need to continue checking other mappings
                relative = Path(f"{{{key}}}").joinpath(relative)
                return relative
        return path

    def platform_path_map(self, path, platform=None):
        """Convert the provided path to one valid for the platform.

        Uses mappings defined in `site.get('platform_path_maps', {})` to convert
        path to the target platform.
        """
        if self.platform == "windows":
            path = PureWindowsPath(path)
        else:
            path = PurePosixPath(path)

        mappings = self.get("platform_path_maps", {})
        # Iterate over all mappings and if applicable, apply each of them
        for mapping in mappings.values():
            src = mapping.get(self.platform)
            dest = mapping.get(platform)
            if path == src:
                # The path is the same as the source, no need to try to resolve
                # a relative path, just convert it to the destination
                path = dest
                continue

            # If path is relative to the current src mapping, replace src with
            # dest to generate the updated path
            try:
                relative = path.relative_to(src)
            except ValueError:
                continue
            else:
                path = dest.joinpath(relative)

        return str(path)

    def run_entry_points_for_group(
        self, group, default=None, entry_points=None, **kwargs
    ):
        """Iterates over `entry_points_for_group` calling the resolved object.

        Args:
            group (str): The name of the group of entry_points to process.
            default (dict, optional): If the entry_point is not defined, return
                the entry points defined by this dictionary. This is the contents
                of the entry_points group, not the entire entry_points dict. For
                example: `{"gui": "hab_gui.cli:gui"}`
            entry_points (dict, optional): Use this dictionary instead of the one
                defined on this Site object.
            **kwargs: Any other kwargs are passed to the loaded entry_point record.
        """
        for ep in self.entry_points_for_group(
            group, default=default, entry_points=entry_points
        ):
            logger.debug(f"Running {group} entry_point: {ep}")
            func = ep.load()
            func(**kwargs)

    def standardize_platform_path_maps(self):
        """Ensure the mappings defined in platform_path_maps are converted to
        the correct PurePath classes."""
        if "platform_path_maps" not in self.frozen_data[self.platform]:
            return
        mappings = self.frozen_data[self.platform]["platform_path_maps"]

        for mapping in mappings.values():
            for platform in mapping:
                if platform == "windows":
                    mapping[platform] = PureWindowsPath(mapping[platform])
                else:
                    mapping[platform] = PurePosixPath(mapping[platform])

    def config_paths(self, config_paths):
        cache = self.cache.config_paths()
        for dirname, path, _ in self.cache.iter_cache_paths(
            "config_paths", config_paths, cache, "*.json"
        ):
            yield dirname, path

    def distro_paths(self, distro_paths):
        cache = self.cache.distro_paths()
        for dirname, path, _ in self.cache.iter_cache_paths(
            "distro_paths", distro_paths, cache, "*/.hab.json"
        ):
            yield dirname, path



class _HabProperty(property):
    """The @property decorator that can be type checked by the `HabMeta` metaclass

    Any properties using this decorator will have their name added to `_properties`.
    Don't use this class directly, use the hab_property decorator instead.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Copy from the class, and store it on the instance. This makes it so we
        # can use isinstance checks against this class.
        cls = type(self)
        self.group = cls.group
        self.process_order = cls.process_order
        self.verbosity = cls.verbosity

    def sort_key(self):
        """Used to provide consistent and controlled when dynamically processing
        _HabProperties.
        """
        return (self.process_order, self.fget.__name__)


def hab_property(verbosity=0, group=1, process_order=100):
    """Decorate this function as a property and configure how it shows up in the dump.

    Args:
        verbosity (int, optional): Specify the minimum verbosity required to show.
            Setting this to None will remove it from the dump. This is useful to
            hide a property that is needed in a super-class but not a subclass.
        group (int, optional): Controls how dump sorts the results. The sort uses
            (group, name) as its sort key, so common group values are sorted
            alphabetically. This should rarely be used and only 0 or 1 should be used
            to reduce the complexity of finding the property you are looking for.
        process_order (int, optional): Fine control over the order properties are
            processed in. The sort uses (process_order, name) as its sort key, so
            common process_order values are sorted alphabetically. This should only
            be needed in specific cases. This is used by ``_HabProperty.sort_key``.
    """
    # Store the requested values to the class so __init__ can copy them into its
    # instance when initialized by the decoration process.
    _HabProperty.group = group
    _HabProperty.process_order = process_order
    _HabProperty.verbosity = verbosity
    return _HabProperty


class HabMeta(type):
    """Scans for HabProperties and adds their name to the `_properties` dict."""

    def __new__(cls, name, bases, dct):  # noqa: B902
        desc = {}
        # Include any _properties that are added by base classes we are inheriting.
        for base in bases:
            if hasattr(base, "_properties"):
                desc.update(base._properties)

        # Add any new _properties defined on this class
        for k, v in dct.items():
            if isinstance(v, _HabProperty):
                desc[k] = v
        dct["_properties"] = desc
        return type.__new__(cls, name, bases, dct)



import json
import os
from contextlib import contextmanager
from pathlib import Path, PurePath

import pytest
from packaging.requirements import Requirement

from hab import Resolver, Site

# Testing both cached and uncached every time adds extra testing time. This env
# var can be used to disable cached testing for local testing.
if os.getenv("HAB_TEST_UNCACHED_ONLY", "0") == "1":
    resolver_tests = ["uncached"]
else:
    resolver_tests = ["uncached", "cached"]


@pytest.fixture(scope="session")
def config_root():
    return Path(__file__).parent


def generate_habcached_site_file(config_root, dest):
    """Returns the path to a site config file generated from `site_main.json`
    configured so it can have a .habcache file generated next to it. The
    config_paths and distro_paths of the site file are hard coded to point to
    the repo's tests directory so it uses the same configs/distros. It also adds
    a `config-root` entry to `platform_path_maps`.
    """
    site_file = Path(dest) / "site.json"
    site_src = config_root / "site_main.json"

    # Load the site_main.json files contents so we can modify it before saving
    # it into the dest for testing.
    data = json.load(site_src.open())
    append = data["append"]

    # Hard code relative_root to the tests folder so it works from
    # a random testing directory without copying all configs/distros.
    for key in ("config_paths", "distro_paths"):
        for i in range(len(append[key])):
            append[key][i] = append[key][i].format(relative_root=site_src.parent)

    # Add platform_path_maps for the pytest directory to simplify testing and
    # test cross-platform support. We need to add all platforms, but this only
    # needs to run on the current platform, so add the same path to all.
    append["platform_path_maps"]["config-root"] = {
        platform: str(config_root) for platform in data["set"]["platforms"]
    }

    with site_file.open("w") as fle:
        json.dump(data, fle, indent=4)

    return site_file


@pytest.fixture(scope="session")
def habcached_site_file(config_root, tmp_path_factory):
    """Generates a site.json file and generates its habcache file.
    This file is stored in a `_cache` directory in the pytest directory.
    This persists for the entire testing session and can be used by other tests
    that need to test hab when it is using a habcache.
    """
    # Create the site file
    shared = tmp_path_factory.mktemp("_cache")
    ret = generate_habcached_site_file(config_root, shared)

    # Generate the habcache file
    site = Site([ret])
    resolver = Resolver(site)
    site.cache.save_cache(resolver, ret)

    return ret


@pytest.fixture
def habcached_resolver(habcached_site_file):
    """Returns a Resolver using a habcache file that was generated for this session.

    See the `habcached_site_file` fixture for details on how the cache is setup.
    For ease of testing the path to the saved habcache file is stored in
    `_test_cache_file` on the returned resolver.
    """
    site = Site([habcached_site_file])
    resolver = Resolver(site)
    # Generate the cache and provide easy access to the habcache file path
    resolver._test_cache_file = site.cache.site_cache_path(habcached_site_file)

    return resolver


@pytest.fixture
def uncached_resolver(config_root):
    """Return a standard testing resolver not using any habcache files."""
    site = Site([config_root / "site_main.json"])
    return Resolver(site=site)


@pytest.fixture(params=resolver_tests)
def resolver(request):
    """Returns a hab.Resolver instance using the site_main.json site config file.

    This is a parameterized fixture that returns both cached and uncached versions
    of the `site_main.json` site configuration. Note the cached version uses a
    copy of it stored in the `_cache0` directory of the pytest temp files. This
    should be used for most tests to ensure that all features are tested, but if
    the test is not affected by caching you can use `uncached_resolver` instead.
    """
    test_map = {"uncached": "uncached_resolver", "cached": "habcached_resolver"}
    return request.getfixturevalue(test_map[request.param])


class Helpers(object):
    """A collection of reusable functions that tests can use."""

    @staticmethod
    def assert_requirements_equal(req, check):
        """Assert that a requirement dict matches a list of requirements.

        Args:
            req (dict): A Requirement dictionary matching the output of
                ``hab.solvers.Solvers.simplify_requirements``.
            check (list): A list of requirement strings. This takes a list
                so writing tests requires less boilerplate.

        Raises:
            AssertionError: If the provided req and check don't exactly match.
        """
        try:
            assert len(req) == len(check)
        except AssertionError:
            # Provide additional information to help debug a failing test. The simple
            # len assert doesn't make it easy to debug a failing test
            print(" Requirement dict ".center(50, "-"))
            print(req)
            print(" Check ".center(50, "-"))
            print(check)
            raise
        for chk in check:
            chk = Requirement(chk)
            assert Helpers.cmp_requirement(req[chk.name], chk)

    @staticmethod
    def check_path_list(paths, checks):
        """Casts the objects in both lists to PurePath objects so they can be
        reliably asserted and differences easily viewed in the pytest output.
        """
        paths = [PurePath(p) for p in paths]
        checks = [PurePath(p) for p in checks]
        assert paths == checks

    @staticmethod
    def cmp_requirement(a, b):
        """Convenience method to check if two Requirement objects are the same.

        Args:
            a (Requirement): The first Requirement to compare
            b (Requirement): The second Requirement to compare

        Returns:
            bool: If a and b represent the same requirement
        """
        return type(a) == type(b) and str(a) == str(b)

    @staticmethod
    @contextmanager
    def reset_environ():
        """Resets the environment variables once the with context exits."""
        old_environ = dict(os.environ)
        try:
            yield
        finally:
            # Restore the original environment variables
            os.environ.clear()
            os.environ.update(old_environ)


@pytest.fixture
def helpers():
    """Expose the Helpers class as a fixture for ease of use in tests."""
    return Helpers



import base64
import errno
import json as _json
import logging.config
import ntpath
import os
import re
import sys
import textwrap
import zlib
from abc import ABC, abstractmethod
from collections import UserDict
from collections.abc import KeysView
from contextlib import contextmanager
from datetime import date, datetime
from pathlib import Path, PurePath

import colorama

# Attempt to use pyjson5 if its installed, this allows us to add comments
# to the various json documents, but is not required. Trailing comments FTW!
try:
    import pyjson5 as json
    from pyjson5 import Json5Exception as _JsonException
except ImportError:
    import json

    class _JsonException(BaseException):
        """Placeholder exception when pyjson5 is not used. Should never be raised"""


colorama.init()

re_windows_single_path = re.compile(r"^([a-zA-Z]:[\\\/][^:;]+)$")
"""A regex that can be used to check if a string is a single windows file path."""


def cygpath(path, spaces=False):
    """Convert a windows path to a cygwin compatible path. For example `c:\\test`
    converts to `/c/test`. This also works for unc file paths.

    Note: This doesn't use cygpath.exe for greater portability. Ie its accessible
    even if the current shell doesn't have access to cygpath.exe. It also doesn't
    require extra suprocess calls.

    Args:
        path (str): The file path to convert.
        spaces (bool, optional): Add a backslash before every non-escaped space.
    """
    path = str(path)

    # Escape spaces and convert any remaining backslashes to forward slashes
    def process_separator(match):
        sep = path[match.start() : match.end()]
        slash_count = sep.count("\\")
        if " " not in sep:
            # It's not a space, simply replace with forward-slash
            return sep.replace("\\", "/")
        # Treat odd numbers of slashes as already escaped not directories.
        elif not spaces or slash_count % 2:
            return sep
        # Add a backslash to escape spaces if enabled
        return sep.replace(" ", "\\ ")

    pattern = (
        # Capture spaces including any leading backslashes to escape
        r"(\\* )"
        # If we can't find any spaces, capture backslashes to convert to forward-slash
        r"|(\\+)"
    )
    path = re.sub(pattern, process_separator, path)

    # Finally, convert `C:\` drive specifier to `/c/`. Unc paths don't need any
    # additional processing, just converting \ to / which was done previously.
    drive, tail = ntpath.splitdrive(path)
    if len(drive) == 2 and drive[1] == ":":
        # It's a drive letter
        path = f"/{drive[0]}{tail}"
    return path


def decode_freeze(txt):
    """Decodes the provided frozen hab string. See `encode_freeze` for
    details on how these strings are encoded. These will start with a version
    identifier `vX:` where X denotes the version it was encoded with."""

    # Extract version information from the string
    try:
        version, txt = txt.split(":", 1)
        if version[0] != "v":
            raise ValueError("Missing v prefix in version information.")
    except ValueError:
        raise ValueError(
            "Missing freeze version information in format `v0:...`"
        ) from None
    try:
        version = int(version[1:])
    except ValueError:
        raise ValueError(f"Version {version[1:]} is not valid.") from None

    data = txt.encode("ascii")
    data = base64.b64decode(txt)
    if version == 1:
        data = data.decode("utf-8")
    elif version == 2:
        data = zlib.decompress(data).decode()
    else:
        return None
    return json.loads(data)


def dump_object(obj, label="", width=80, flat_list=False, color=False):
    """Recursively convert python objects into a human readable table string.

    Args:
        obj: The object to convert. This is called recursively for dicts and
            lists, other objects are simply cast as a string.
        label (str, optional): Prefix the first line with this text, and any
            additional lines are prefixed with white space of the same length.
            Include any desired formatting including white space.
        width (int, optional): The desired width for wrapping. The output may
            exceed this value, but it will attempt to respect it.
        flat_list (bool, optional): By default for lists, it will attempt to
            store multiple list items on a single line, but if that ends up
            exceeding width, it will switch to a line per item, this is useful
            for path output. Setting this to True will force it to put as many
            items on a single line as possible instead. It will not allow an
            item to be broken into multiple lines.
        color (bool, optional): Use ANSI escape character sequences to colorize
            the output of the text.
    """
    pad = " " * len(label)
    if label:
        width = width - len(label)
        if width < 10:
            width = 10
        if color:
            label = f"{colorama.Fore.GREEN}{label}{colorama.Style.RESET_ALL}"

    if isinstance(obj, (list, KeysView)):
        rows = []
        obj = [
            dump_object(o, width=width, flat_list=flat_list, color=color) for o in obj
        ]
        if flat_list:
            # combine as many list items as possible onto each line
            # Use non-breaking spaces to prevent word wrap for each item.
            obj = textwrap.wrap(
                " ".join([x.replace(" ", "u\00A0") for x in obj]),
                width=width,
                break_long_words=False,
            )
            # Remove those pesky non-breaking spaces
            obj = [x.replace("u\00A0", " ") for x in obj]
            one_row = ", ".join(obj)
            multi_row = len(obj) > 1
        else:
            # Determine if we can store all of this on a single line within
            # the requested width
            one_row = ", ".join(obj)
            multi_row = len(one_row) > width

        if multi_row:
            rows.append(f"{label}{obj[0]}")
            rows.extend([f"{pad}{o}" for o in obj[1:]])
        else:
            rows = [f"{label}{one_row}"]

        return "\n".join(rows)
    elif isinstance(obj, (dict, UserDict)):
        rows = []
        lbl = label
        for k, v in sorted(obj.items()):
            rows.append(
                dump_object(
                    v,
                    label=f"{lbl}{k}:  ",
                    width=width,
                    flat_list=flat_list,
                    color=color,
                )
            )
            lbl = pad
        return "\n".join(rows)
    elif isinstance(obj, PurePath):
        return f"{label}{obj}"
    elif hasattr(obj, "name"):
        # Likely HabBase objects
        return f"{label}{obj.name}"
    # Simply convert any other objects to strings
    return f"{label}{obj}"


def dump_title(title, body, color=False):
    """Returns the body with title and wrapped in a header and footer line

    Args:
        title (str): The title to add before the header line.
        body (str): The text to be wrapped in the header and footer lines.
        color (bool, optional): Use ANSI escape character sequences to colorize
            the output of the text.
    """
    # Find the max width of the body and title to add the line
    width = len(max(body.split("\n"), key=len))
    width = max(len(title), width)
    if color:
        title = f"{colorama.Fore.GREEN}{title}{colorama.Style.RESET_ALL}"
    return f"{title}\n{'-'*width}\n{body}\n{'-'*width}"


def dumps_json(data, **kwargs):
    """Consistent method for calling json.dumps that handles encoding custom
    data objects like NotSet.

    Pyjson5's dump is not as fully featured as python's json, so this ensures
    consistent dumps output as python's json module has more features than
    pyjson5. For example pyjson5 doesn't support indent."""
    kwargs.setdefault("cls", HabJsonEncoder)
    return _json.dumps(data, **kwargs)


def encode_freeze(data, version=None, site=None):
    """Encodes the provided data object in json. This string is stored on the
    "HAB_FREEZE" environment variable when run by the cli.

    Encoded freeze version data is a string with a version prefix matching the
    regex `v(?P<ver>\\d+):(?P<freeze>.+)`.

    Version 1: Encodes data to a json string that is then encoded with base64.
    Version 2: Encodes data to a json string that is then compressed with zlib,
        and finally encoded with base64.
    None is returned if this version number is not supported.

    Args:
        data: The data to freeze.
        version (int, optional): The version to encode the freeze with.
            If None is passed, then the default version encoding is used.
        site (hab.site.Site, optional): If version is not specified, then
            attempt to get version from `site.get("freeze_version")`.
    """
    if version is None and site:
        version = site.get("freeze_version")

    if version is None:
        # The current default is 2. Using None makes it easy to control this
        # with a site configuration.
        version = 2

    data = dumps_json(data)
    data = data.encode("utf-8")
    if version == 1:
        data = base64.b64encode(data)
    elif version == 2:
        data = zlib.compress(data)
        data = base64.b64encode(data)
    else:
        return None

    return f'v{version}:{data.decode("utf-8")}'


class HabJsonEncoder(_json.JSONEncoder):
    """JsonEncoder class that handles non-supported objects like hab.NotSet."""

    def default(self, obj):
        if obj is NotSet:
            # Convert NotSet to None for json storage
            return None
        # Handle date and datetime conversion
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()

        # Let the base class default method raise the TypeError
        return _json.JSONEncoder.default(self, obj)


def load_json_file(filename):
    """Open and parse a json file. If a parsing error happens the file path is
    added to the exception to allow for easier debugging.

    Args:
        filename (pathlib.Path): A existing file path.

    Returns:
        The data stored in the json file.

    Raises:
        FileNotFoundError: If filename is not pointing to a file that actually exists.
        pyjson5.Json5Exception: If using pyjson5, the error raised due to invalid json.
        ValueError: If not using pyjson5, the error raised due to invalid json.
    """
    if not filename.is_file():
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(filename))

    with filename.open() as fle:
        try:
            data = json.load(fle)
        # Include the filename in the traceback to make debugging easier
        except _JsonException as e:
            # pyjson5 is installed
            e.result["filename"] = str(filename)
            raise e.with_traceback(sys.exc_info()[2]) from None
        except ValueError as e:
            # Using python's native json parser
            msg = f'{e} Filename("{filename}")'
            raise type(e)(msg, e.doc, e.pos).with_traceback(sys.exc_info()[2]) from None
    return data


def natural_sort(ls, key=None):
    """Sort a list in a more natural way by treating contiguous integers as a
    single number instead of processing each number individually. This function
    understands that 10 is larger than 1. It also ignores case.

    Source: http://blog.codinghorror.com/sorting-for-humans-natural-sort-order
    """

    def convert(text):
        return int(text) if text.isdigit() else text.lower()

    if key is None:

        def key(text):
            return text

    def alphanum_key(a_key):
        return [convert(c) for c in re.split(r"([0-9]+)", key(a_key))]

    return sorted(ls, key=alphanum_key)


class NotSet(object):
    """The data for this property is not currently set."""

    def __bool__(self):
        """NotSet should be treated as False when booled Python 3"""
        return False

    def __str__(self):
        return "NotSet"

    def __copy__(self):
        """Does not return a copy of this object, NotSet is intended to be a
        singleton so it does not copy itself."""
        return self

    def __deepcopy__(self, memo):
        """Does not return a copy of this object, NotSet is intended to be a
        singleton so it does not copy itself."""
        return self

    def __reduce__(self):
        """Enable pickling of NotSet."""
        return type(self).__qualname__


# Make this a singleton so it works like a boolean False for if statements.
NotSet = NotSet()


def path_forward_slash(path):
    """Converts a Path object into a string with forward slashes"""
    return str(path).replace("\\", "/")


@contextmanager
def verbosity_filter(resolver, verbosity, target=None):
    """Change the verbosity settings of a hab.Resolver while inside this with
    context.

    Args:
        resolver (hab.Resolver): The resolver to change verbosity settings on.
        verbosity (int): Change the verbosity setting to this value. If None is
            passed, all results are be shown without any filtering.
        target (str, optional): Change the verbosity target, ignored if None.
    """
    # Backup the current values
    current_target = resolver._verbosity_target
    current_value = resolver._verbosity_value

    # Change to the requested values
    if target is not None:
        resolver._verbosity_target = target
    resolver._verbosity_value = verbosity

    try:
        yield
    finally:
        # Restore the original values
        resolver._verbosity_target = current_target
        resolver._verbosity_value = current_value


class BasePlatform(ABC):
    """Subclasses of BasePlatform are the interface hab uses to handle cross
    platform code.

    The current platform is accessible via hab.utils.Platform which is set by
    calling `BasePlatform.get_platform()`. To change the current platform, set
    the subclass to Platform. `hab.utils.Platform = hab.utils.WinPlatform`
    """

    _default_ext = ".sh"
    _name = None
    _sep = ":"

    @classmethod
    def configure_logging(cls, filename=None):
        """Update the logging configuration with the contents of this json file
        if exists.

        See https://docs.python.org/3/library/logging.config.html#dictionary-schema-details
        for details on how to construct this file.

        You will most likely want to enable incremental so it doesn't fully reset
        the logging basicConfig.

        Example:
            {"incremental": True,
             "loggers": {"": {"level": 30}, "hab.parsers": {"level": 10}},
             "version": 1}
        """
        if filename is None:
            filename = cls.user_prefs_filename(".hab_logging_prefs.json")

        if not filename.exists():
            return False

        with filename.open() as fle:
            cfg = json.load(fle)
            logging.config.dictConfig(cfg)
            return True

    @classmethod
    @abstractmethod
    def check_name(cls, name):
        """Checks if the provided name is valid for this class"""

    @classmethod
    def collapse_paths(cls, paths, ext=None, key=None):
        """Converts a list of paths into a string compatible with the os.

        Args:
            paths: A list of paths that have str called on each of them. If
                a string is passed, it is returned un-modified.
            ext (str, optional): Used to apply special formatting rules based on
                the shell script being written.
            key (str, optional): Used to apply special formatting rules based on
                the environment variable this being used to process.
        """
        if isinstance(paths, str):
            return paths
        return cls.pathsep(ext=ext).join([str(p) for p in paths])

    @classmethod
    def default_ext(cls):
        """Returns the default file extension used on this platform."""
        return cls._default_ext

    @classmethod
    def expand_paths(cls, paths):
        """Converts path strings separated by ``cls.pathsep()`` and lists into
        a list containing Path objects.
        """
        if isinstance(paths, str):
            return [Path(p) for p in paths.split(cls.pathsep())]
        return [Path(p) for p in paths]

    @staticmethod
    def get_platform(name=None):
        """Returns the subclass matching the requested platform name. This can
        be the value returned by sys.platform, BasePlatform.system(), or a name
        returned by a subclass. See `check_name` for details.
        """
        if name is None:
            name = BasePlatform.system()
        # Note: This is a staticmethod because it should always look at its
        # children not the current class's children
        for c in BasePlatform.__subclasses__():
            if c.check_name(name):
                return c

    @classmethod
    def name(cls):
        """The hab name for this platform."""
        return cls._name

    @classmethod
    def path_split(cls, path, pathsep=None):
        """Split a string by pathsep unless that path is a single windows path.
        This is used to detect an windows file path on linux which uses `:` for
        path separator and conflicts with the drive letter specification.

        Args:
            path (str): The string to split.
            pathsep (str, optional): If not specified, `os.pathsep` is used.

        Returns:
            list: A list of individual paths.
        """
        if pathsep is None:
            pathsep = cls.pathsep()

        # If on linux we need to resolve a windows path we can't just use
        # `os.path.pathsep`, check if this is a single windows file path and
        # return it as a list.
        if pathsep == ":" and re_windows_single_path.match(path):
            return [path]

        return path.split(pathsep)

    @classmethod
    def pathsep(cls, ext=None):
        """The path separator used by this platform."""
        return cls._sep

    @classmethod
    def system(cls):
        """Returns the current operating system as `windows`, `osx` or `linux`."""
        if sys.platform == "darwin":
            return "osx"
        if sys.platform == "win32":
            return "windows"
        return "linux"

    @classmethod
    def user_prefs_filename(cls, filename=".hab_user_prefs.json"):
        """Returns the filename that contains the hab user preferences."""
        return Path.home() / filename


class WinPlatform(BasePlatform):
    _default_ext = ".bat"
    _name = "windows"
    _sep = ";"

    @classmethod
    def check_name(cls, name):
        return name in ("win32", "windows")

    @classmethod
    def collapse_paths(cls, paths, ext=None, key=None):
        """Converts a list of paths into a string compatible with the os.

        If ext is `.sh` and key is `PATH` the paths are converted using cygpath
        and `:` is used for pathsep instead of `;`. This is necessary due to cygwin
        converting the PATH env var to linux. It apparently only does this for
        PATH, so other env var's simply have `str` called on them and uses `;`
        for pathsep. See https://cygwin.com/cygwin-ug-net/setup-env.html

        Args:
            paths: A list of paths that have str called on each of them. If
                a string is passed, it is returned un-modified.
            ext (str, optional): Used to apply special formatting rules based on
                the shell script being written.
            key (str, optional): Used to apply special formatting rules based on
                the environment variable this being used to process.
        """
        if isinstance(paths, str):
            return paths
        if ext in (".sh", "") and key == "PATH":
            paths = [cygpath(p) for p in paths]
        else:
            paths = [str(p) for p in paths]
        return cls.pathsep(ext=ext, key=key).join(paths)

    @classmethod
    def pathsep(cls, ext=None, key=None):
        """The path separator used by this platform."""
        if ext in (".sh", "") and key == "PATH":
            # For shwin scripts we should use linux style pathsep
            return ":"
        return cls._sep

    @classmethod
    def user_prefs_filename(cls, filename=".hab_user_prefs.json"):
        """Returns the filename that contains the hab user preferences."""
        return Path(os.path.expandvars("$LOCALAPPDATA")) / filename


class LinuxPlatform(BasePlatform):
    _name = "linux"

    @classmethod
    def check_name(cls, name):
        return name.startswith("linux")


class OsxPlatform(BasePlatform):
    _name = "osx"

    @classmethod
    def check_name(cls, name):
        return name in ("darwin", "osx")


Platform = BasePlatform.get_platform()
"""BasePlatform: Hab uses this variable to handle any platform specific operations
and to know what the current platform is. This allows for centralization of
platform specific code. It also allows the testing suite to switch to other
platforms to prevent the need for a developer to test their changes on each
platform individually, though the CI testing should still run on all platforms.
"""



import sys
from pathlib import Path, PurePosixPath, PureWindowsPath

import pytest
from colorama import Fore, Style

from hab import Resolver, Site, utils
from hab.cache import Cache


def test_environment_variables(config_root, monkeypatch):
    paths = [config_root / "site_main.json"]
    monkeypatch.setenv("HAB_PATHS", utils.Platform.collapse_paths(paths))
    site = Site()
    assert site.paths == paths

    paths.append(config_root / "site_override.json")
    monkeypatch.setenv("HAB_PATHS", utils.Platform.collapse_paths(paths))
    site = Site()
    assert site.paths == paths


class TestMultipleSites:
    """Check that various combinations of site json files results in the correct
    merged site. The rules are:

    1. The left most site configuration takes precedence for a given item.
    2. For prepend/append operations on lists, the left site file's paths will
    placed on the outside of the the right site file's paths.
    3. For `platform_path_maps`, only the first key is kept and any duplicates
    are discarded.
    """

    def host_left(self):
        return {
            "linux": PurePosixPath("host-linux_left"),
            "windows": PureWindowsPath("host-windows_left"),
        }

    def host_middle(self):
        return {
            "linux": PurePosixPath("host-linux_middle"),
            "windows": PureWindowsPath("host-windows_middle"),
        }

    def host_right(self):
        return {
            "linux": PurePosixPath("host-linux_right"),
            "windows": PureWindowsPath("host-windows_right"),
        }

    def mid(self):
        return {
            "linux": PurePosixPath("mid-linux_middle"),
            "windows": PureWindowsPath("mid-windows_middle"),
        }

    def net(self):
        return {
            "linux": PurePosixPath("net-linux_right"),
            "windows": PureWindowsPath("net-windows_right"),
        }

    def shared_left(self):
        return {
            "linux": PurePosixPath("shared-linux_left"),
            "windows": PureWindowsPath("shared-windows_left"),
        }

    def test_left(self, config_root):
        """Check that site_left.json returns the correct results."""
        paths = [config_root / "site" / "site_left.json"]
        site = Site(paths)

        assert site.get("set_value") == ["left"]
        assert site.get("test_paths") == ["left_prepend", "left_append"]

        check = {"host": self.host_left(), "shared": self.shared_left()}

        assert site.get("platform_path_maps") == check

    def test_middle(self, config_root):
        """Check that site_middle.json returns the correct results."""
        paths = [config_root / "site" / "site_middle.json"]
        site = Site(paths)

        assert site.get("set_value") == ["middle"]
        assert site.get("test_paths") == ["middle_prepend", "middle_append"]

        check = {"host": self.host_middle(), "mid": self.mid()}

        assert site.get("platform_path_maps") == check

    def test_right(self, config_root):
        """Check that site_right.json returns the correct results."""
        paths = [config_root / "site" / "site_right.json"]
        site = Site(paths)

        assert site.get("set_value") == ["right"]
        assert site.get("test_paths") == ["right_prepend", "right_append"]

        check = {"host": self.host_right(), "net": self.net()}

        assert site.get("platform_path_maps") == check

    def test_left_right(self, config_root):
        """Check that site_left.json and site_right.json are merged correctly."""
        paths = [
            config_root / "site" / "site_left.json",
            config_root / "site" / "site_right.json",
        ]
        site = Site(paths)

        assert site.get("set_value") == ["left"]
        assert site.get("test_paths") == [
            "left_prepend",
            "right_prepend",
            "right_append",
            "left_append",
        ]

        check = {
            "host": self.host_left(),
            "net": self.net(),
            "shared": self.shared_left(),
        }

        assert site.get("platform_path_maps") == check

    def test_right_left(self, config_root):
        """Check that reversing site_left.json and site_right.json get merged
        correctly."""
        paths = [
            config_root / "site" / "site_right.json",
            config_root / "site" / "site_left.json",
        ]
        site = Site(paths)

        assert site.get("set_value") == ["right"]
        assert site.get("test_paths") == [
            "right_prepend",
            "left_prepend",
            "left_append",
            "right_append",
        ]

        check = {
            "host": self.host_right(),
            "net": self.net(),
            "shared": self.shared_left(),
        }

        assert site.get("platform_path_maps") == check

    def test_left_middle_right(self, config_root):
        """Check that more than 2 site json files are merged correctly."""
        paths = [
            config_root / "site" / "site_left.json",
            config_root / "site" / "site_middle.json",
            config_root / "site" / "site_right.json",
        ]
        site = Site(paths)

        assert site.get("set_value") == ["left"]
        assert site.get("test_paths") == [
            "left_prepend",
            "middle_prepend",
            "right_prepend",
            "right_append",
            "middle_append",
            "left_append",
        ]

        check = {
            "host": self.host_left(),
            "mid": self.mid(),
            "net": self.net(),
            "shared": self.shared_left(),
        }

        assert site.get("platform_path_maps") == check


class TestResolvePaths:
    def test_path(self, config_root):
        # Check that the correct values are resolved when processing a single result
        paths = [config_root / "site_main.json"]
        site = Site(paths)
        assert site.get("generic_value") is False
        assert "override" not in site
        assert site.get("filename") == ["site_main.json"]

    def test_paths(self, config_root, helpers):
        # Check that values specified by additional files overwrite the previous values
        paths = [config_root / "site_override.json", config_root / "site_main.json"]
        site = Site(paths)
        assert site.get("generic_value") is True
        assert site.get("override") == ["site_override.json"]
        assert site.get("filename") == ["site_override.json"]
        assert site.get("platforms") == ["windows", "linux"]

        # Check that the paths defined in multiple site files are correctly added
        assert len(site.get("config_paths")) == 1
        helpers.check_path_list(
            site.get("config_paths"), [config_root / "configs" / "*"]
        )
        assert len(site.get("distro_paths")) == 2
        helpers.check_path_list(
            site.get("distro_paths"),
            (
                config_root / "distros" / "*",
                config_root / "duplicates" / "distros_1" / "*",
            ),
        )

    def test_paths_reversed(self, config_root, helpers):
        # Check that values specified by additional files overwrite the previous values
        paths = [config_root / "site_main.json", config_root / "site_override.json"]
        site = Site(paths)
        assert site.get("generic_value") is False
        assert site.get("override") == ["site_override.json"]
        assert site.get("filename") == ["site_main.json"]

        # Check that the paths defined in multiple site files are correctly added
        assert len(site.get("config_paths")) == 1
        helpers.check_path_list(
            site.get("config_paths"), [config_root / "configs" / "*"]
        )
        assert len(site.get("distro_paths")) == 2
        helpers.check_path_list(
            site.get("distro_paths"),
            (
                config_root / "duplicates" / "distros_1" / "*",
                config_root / "distros" / "*",
            ),
        )


def test_path_in_raise(config_root):
    paths = [config_root / "site_main.json", config_root / "missing_file.json"]
    with pytest.raises(FileNotFoundError) as excinfo:
        Site(paths)
    assert "No such file or directory:" in str(excinfo.value)
    assert "missing_file.json" in str(excinfo.value)


def test_dump(config_root):
    """utils.dump_object are checked pretty well in test_parsing, here we test
    the colorization settings and ensuring that the desired results are listed
    """
    checks = (
        "{green}Dump of Site{reset}\n",
        "{green}ignored_distros:  {reset}release, pre",
    )

    paths = [config_root / "site_main.json"]
    site = Site(paths)
    assert site.get("colorize") is None

    result = site.dump()
    for check in checks:
        assert check.format(green=Fore.GREEN, reset=Style.RESET_ALL) in result

    paths = [config_root / "site_override.json"]
    site = Site(paths)
    assert site.get("colorize") is False

    result = site.dump()
    for check in checks:
        assert check.format(green="", reset="") in result


def test_dump_cached(config_root, habcached_site_file):
    """Test that cached indicators are shown properly for site dump based on
    verbosity setting."""

    # Create the hab setup with caching only on one of the two site files
    other_site = config_root / "site_os_specific.json"
    site = Site([habcached_site_file, other_site])
    resolver = Resolver(site)
    site.cache.save_cache(resolver, habcached_site_file)

    # Build a check string to verify that the dump is correctly formatted
    # Note: To simplify the check_template for testing we will force the dump
    # to a smaller width to ensure it always wraps the file paths.
    platform = utils.Platform.name()
    check_template = (
        f"{{green}}HAB_PATHS:  {{reset}}{habcached_site_file}{{cached}}",
        f"            {other_site}",
        f"{{green}}config_paths:  {{reset}}config\\path\\{platform}",
        f"               {config_root}\\configs\\*{{cached}}",
        f"{{green}}distro_paths:  {{reset}}distro\\path\\{platform}",
        f"               {config_root}\\distros\\*{{cached}}",
    )
    check_template = "\n".join(check_template)
    colors = {
        "green": Fore.GREEN,
        "reset": Style.RESET_ALL,
    }
    if platform != "windows":
        check_template = check_template.replace("\\", "/")

    # With color enabled:
    # No verbosity, should not show cached status
    assert site.get("colorize") is None
    result = site.dump(width=60)
    check = check_template.format(cached="", **colors)
    assert check in result

    # verbosity enabled, should show cached status
    result = site.dump(verbosity=1, width=60)
    check = check_template.format(
        cached=f" {Fore.YELLOW}(cached){Style.RESET_ALL}", **colors
    )
    assert check in result

    # Disable Color:
    site["colorize"] = False
    assert site.get("colorize") is False

    # No verbosity, should not show cached status
    result = site.dump(width=60)
    check = check_template.format(cached="", green="", reset="")
    assert check in result

    # verbosity enabled, should show cached status
    result = site.dump(verbosity=1, width=60)
    check = check_template.format(cached=" (cached)", green="", reset="")
    assert check in result


class TestOsSpecific:
    def test_linux(self, monkeypatch, config_root):
        """Check that if "os_specific" is set to true, only vars for the current
        os are resolved."""
        # Simulate running on a linux platform.
        monkeypatch.setattr(utils, "Platform", utils.LinuxPlatform)

        paths = [config_root / "site_os_specific.json"]
        site = Site(paths)

        assert site.get("config_paths") == [Path("config/path/linux")]
        assert site.get("distro_paths") == [Path("distro/path/linux")]
        assert site.get("platforms") == ["windows", "linux"]

    def test_osx(self, monkeypatch, config_root):
        """Check that if "os_specific" is set to true, only vars for the current
        os are resolved."""
        # Simulate running on a osx platform.
        monkeypatch.setattr(utils, "Platform", utils.OsxPlatform)

        paths = [config_root / "site_os_specific.json"]
        site = Site(paths)

        assert site.get("config_paths") == [Path("config/path/osx")]
        assert site.get("distro_paths") == [Path("distro/path/osx")]
        assert site.get("platforms") == ["osx", "linux"]

    def test_win(self, monkeypatch, config_root):
        """Check that if "os_specific" is set to true, only vars for the current
        os are resolved."""
        # Simulate running on a windows platform
        monkeypatch.setattr(utils, "Platform", utils.WinPlatform)

        paths = [config_root / "site_os_specific.json"]
        site = Site(paths)

        assert site.get("config_paths") == [Path("config\\path\\windows")]
        assert site.get("distro_paths") == [Path("distro\\path\\windows")]
        assert site.get("platforms") == ["windows", "osx"]


class TestPlatformPathMap:
    def test_linux(self, monkeypatch, config_root):
        """For linux check that various inputs are correctly processed."""
        monkeypatch.setattr(utils, "Platform", utils.LinuxPlatform)
        site = Site([config_root / "site_main.json"])

        # Check exact path matches are translated
        out = site.platform_path_map("/usr/local/host/root", platform="linux")
        assert out == "/usr/local/host/root"
        out = site.platform_path_map("/usr/local/host/root/extra", platform="linux")
        assert out == "/usr/local/host/root/extra"
        out = site.platform_path_key("/usr/local/host/root", platform="linux")
        assert out.as_posix() == "{host-root}"
        out = site.platform_path_key("/usr/local/host/root/extra", platform="linux")
        assert out.as_posix() == "{host-root}/extra"

        out = site.platform_path_map("/usr/local/host/root", platform="osx")
        assert out == "/usr/local/osx/host/root"
        out = site.platform_path_map("/usr/local/host/root/extra", platform="osx")
        assert out == "/usr/local/osx/host/root/extra"
        out = site.platform_path_key("/usr/local/host/root", platform="osx")
        assert out.as_posix() == "{host-root}"
        out = site.platform_path_key("/usr/local/host/root/extra", platform="osx")
        assert out.as_posix() == "{host-root}/extra"

        out = site.platform_path_map("/usr/local/host/root", platform="windows")
        assert out == r"c:\host\root"
        out = site.platform_path_map("/usr/local/host/root/extra", platform="windows")
        assert out == r"c:\host\root\extra"
        out = site.platform_path_key("/usr/local/host/root", platform="windows")
        assert out.as_posix() == "{host-root}"
        out = site.platform_path_key("/usr/local/host/root/extra", platform="windows")
        assert out.as_posix() == "{host-root}/extra"

    def test_win(self, monkeypatch, config_root):
        """For windows check that various inputs are correctly processed."""
        monkeypatch.setattr(utils, "Platform", utils.WinPlatform)
        site = Site([config_root / "site_main.json"])

        # Check exact path matches are translated
        out = site.platform_path_map(r"c:\host\root", platform="linux")
        assert out == "/usr/local/host/root"
        out = site.platform_path_map(r"c:\host/root/extra", platform="linux")
        assert out == "/usr/local/host/root/extra"
        out = site.platform_path_map("c:/host/root", platform="linux")
        assert out == "/usr/local/host/root"
        out = site.platform_path_key(r"c:\host\root", platform="linux")
        assert out.as_posix() == "{host-root}"
        out = site.platform_path_key(r"c:\host\root\extra", platform="linux")
        assert out.as_posix() == "{host-root}/extra"

        out = site.platform_path_map(r"c:\host\root", platform="osx")
        assert out == "/usr/local/osx/host/root"
        out = site.platform_path_map(r"c:\host\root\extra", platform="osx")
        assert out == "/usr/local/osx/host/root/extra"
        out = site.platform_path_map("c:/host/root", platform="osx")
        assert out == "/usr/local/osx/host/root"
        out = site.platform_path_key(r"c:\host\root", platform="osx")
        assert out.as_posix() == "{host-root}"
        out = site.platform_path_key(r"c:\host\root\extra", platform="osx")
        assert out.as_posix() == "{host-root}/extra"

        out = site.platform_path_map(r"c:\host\root", platform="windows")
        assert out == r"c:\host\root"
        out = site.platform_path_map(r"c:\host\root\extra", platform="windows")
        assert out == r"c:\host\root\extra"
        out = site.platform_path_map("c:/host/root", platform="windows")
        assert out == r"c:\host\root"
        out = site.platform_path_key(r"c:\host\root", platform="windows")
        assert out.as_posix() == "{host-root}"
        out = site.platform_path_key(r"c:\host\root\extra", platform="windows")
        assert out.as_posix() == "{host-root}/extra"

    def test_unset_variables(self, config_root):
        """Don't modify variables that are not specified in platform_path_map"""
        site = Site([config_root / "site_main.json"])
        out = site.platform_path_key("{unset-variable}/is/not/modified")
        assert out.as_posix() == "{unset-variable}/is/not/modified"


class TestPlatformPathMapDict:
    def assert_main(self, site):
        """Used by several tests to assert site_main.json's "network-mount"
        settings are in use.
        """
        assert site["platform_path_maps"]["network-mount"]["linux"] == PurePosixPath(
            "/mnt/shared_resources"
        )
        assert site["platform_path_maps"]["network-mount"]["osx"] == PurePosixPath(
            "/mnt/osx/shared_resources"
        )
        assert site["platform_path_maps"]["network-mount"][
            "windows"
        ] == PureWindowsPath("\\\\example\\shared_resources")

    def assert_override(self, site):
        """Used by several tests to assert site_override.json's "network-mount"
        settings are in use.
        """
        assert site["platform_path_maps"]["network-mount"]["linux"] == PurePosixPath(
            "/mnt/work/shared_resources"
        )
        assert site["platform_path_maps"]["network-mount"]["osx"] == PurePosixPath(
            "/mnt/work/osx/shared_resources"
        )
        assert site["platform_path_maps"]["network-mount"][
            "windows"
        ] == PureWindowsPath("g:\\work\\shared_resources")

    def test_main(self, config_root):
        """Test that all platform_path_map settings in site_main.json were found."""
        paths = [config_root / "site_main.json"]
        site = Site(paths)

        assert site["platform_path_maps"]["host-root"]["linux"] == PurePosixPath(
            "/usr/local/host/root"
        )
        assert site["platform_path_maps"]["host-root"]["osx"] == PurePosixPath(
            "/usr/local/osx/host/root"
        )
        assert site["platform_path_maps"]["host-root"]["windows"] == PureWindowsPath(
            "c:\\host\\root"
        )

        self.assert_main(site)

    def test_override(self, config_root):
        """Test that all platform_path_map settings in site_override.json were found."""
        paths = [config_root / "site_override.json"]
        site = Site(paths)

        # host-root is not specified in this site file.
        assert "host-root" not in site["platform_path_maps"]

        self.assert_override(site)

    def test_merged(self, config_root):
        """Test that the correct platform_path_map settings are resolved when using
        both site_main.json and site_override.json."""
        paths = [config_root / "site_main.json", config_root / "site_override.json"]
        site = Site(paths)

        # This is only specified in site_main.json
        assert site["platform_path_maps"]["host-root"]["linux"] == PurePosixPath(
            "/usr/local/host/root"
        )
        assert site["platform_path_maps"]["host-root"]["osx"] == PurePosixPath(
            "/usr/local/osx/host/root"
        )
        assert site["platform_path_maps"]["host-root"]["windows"] == PureWindowsPath(
            "c:\\host\\root"
        )

        # The right-most site file's settings are loaded when both have the same keys
        self.assert_main(site)

    def test_reversed(self, config_root):
        """Test that the correct platform_path_map settings are resolved when using
        both site_main.json and site_override.json."""
        paths = [config_root / "site_override.json", config_root / "site_main.json"]
        site = Site(paths)

        # This is only specified in site_main.json
        assert site["platform_path_maps"]["host-root"]["linux"] == PurePosixPath(
            "/usr/local/host/root"
        )
        assert site["platform_path_maps"]["host-root"]["osx"] == PurePosixPath(
            "/usr/local/osx/host/root"
        )
        assert site["platform_path_maps"]["host-root"]["windows"] == PureWindowsPath(
            "c:\\host\\root"
        )

        # The right-most site file's settings are loaded when both have the same keys
        self.assert_override(site)


class TestEntryPoints:
    def test_empty_site(self, config_root):
        """Test a site not defining any entry points for `hab.cli`."""
        site = Site([config_root / "site_main.json"])
        entry_points = site.entry_points_for_group("hab.cli")
        assert len(entry_points) == 0

    def test_default(self, config_root):
        """Test a site not defining any entry points for `hab.cli`."""
        site = Site([config_root / "site_main.json"])
        entry_points = site.entry_points_for_group(
            "hab.cli", default={"test": "case:func"}
        )
        assert len(entry_points) == 1

        # Test that the `test-gui` `hab.cli` entry point is handled correctly
        ep = entry_points[0]
        assert ep.name == "test"
        assert ep.group == "hab.cli"
        assert ep.value == "case:func"

    @pytest.mark.parametrize(
        # Note: The default for omit_none is True
        "site_files,import_name,fname,omit_none",
        (
            (["site/site_entry_point_a.json"], "hab_test_entry_points", "gui", True),
            (
                ["site/site_entry_point_b.json", "site/site_entry_point_a.json"],
                "hab_test_entry_points",
                "gui_alt",
                True,
            ),
            (
                ["site/site_entry_point_a.json", "site/site_entry_point_b.json"],
                "hab_test_entry_points",
                "gui",
                True,
            ),
            # Tests handling an entry_point value of None
            (
                # None value is ignored due to order
                ["site/site_entry_point_a.json", "site/site_entry_point_c.json"],
                "hab_test_entry_points",
                "gui",
                True,
            ),
            (
                # None value is used, but ignored due to omit_none setting
                ["site/site_entry_point_c.json", "site/site_entry_point_a.json"],
                None,
                None,
                True,
            ),
            (
                # None value is used, but still returned due to omit_none setting
                ["site/site_entry_point_c.json", "site/site_entry_point_a.json"],
                None,
                None,
                False,
            ),
        ),
    )
    def test_site_cli(self, config_root, site_files, import_name, fname, omit_none):
        """Test a site defining an entry point for `hab.cli`, possibly multiple times."""
        site = Site([config_root / f for f in site_files])
        entry_points = site.entry_points_for_group("hab.cli", omit_none=omit_none)

        if import_name is None and omit_none is True:
            assert len(entry_points) == 0
            # Nothing else to test if the value is null
            return

        assert len(entry_points) == 1

        # Test that the `test-gui` `hab.cli` entry point is handled correctly
        ep = entry_points[0]
        assert ep.name == "test-gui"
        assert ep.group == "hab.cli"
        if omit_none is False:
            assert ep.value is None
            # Noting else to test, we can't load a value of None.
            return
        else:
            assert ep.value == f"{import_name}:{fname}"

        # Load the module's function
        funct = ep.load()

        # The module has now been imported and the correct function was loaded
        assert funct is getattr(sys.modules[import_name], fname)
        with pytest.raises(
            NotImplementedError, match=rf"{import_name}\.{fname} called successfully"
        ):
            funct()

    @pytest.mark.parametrize(
        "site_file,except_match",
        (
            (
                "cfg_reduce_env.json",
                "hab_test_entry_points.cfg_reduce_env called successfully",
            ),
            (
                "cfg_reduce_finalize.json",
                "hab_test_entry_points.cfg_reduce_finalize called successfully",
            ),
            (
                "cfg_uri_validate.json",
                "hab_test_entry_points.uri_validate_error called successfully",
            ),
        ),
    )
    def test_called_by_resolve(self, config_root, site_file, except_match):
        """Test that site defined entry_points are called.

        This expects that the entry point will raise a `NotImplementedError` with
        a specific message. This requires that each test has its own site json
        file enabling that specific entry_point. See `tests/site/eps/README.md`.
        """
        site = Site(
            [
                config_root / "site" / "eps" / site_file,
                config_root / "site_main.json",
            ]
        )
        resolver = Resolver(site=site)

        # The module has now been imported and the correct function was loaded
        with pytest.raises(NotImplementedError, match=except_match):
            resolver.resolve("default")

    def test_site_add_paths_non_recursive(self, config_root):
        """Checks that the `hab.site.add_paths` entry_point is respected for
        file paths passed to the paths argument of Site. Also test that the
        entry_point is ignored when processing these dynamically added paths.
        """
        site = Site(
            [
                config_root / "site" / "eps" / "site_add_paths.json",
            ]
        )

        # Check that static and dynamic paths were added in the correct order.
        assert len(site.paths) == 3
        assert site.paths[0].name == "site_add_paths_a.json"
        assert site.paths[1].name == "site_add_paths_b.json"
        assert site.paths[2].name == "site_add_paths.json"

        # Check which "set" value was resolved by the end. To correctly process
        # the list returned by the entry_points are processed in reverse order
        assert site["test_data"] == ["site_add_paths_a.json"]

    def test_site_add_paths_multiple(self, config_root):
        """Checks that multiple `hab.site.add_paths` entry_points are processed
        when not added dynamically."""
        site = Site(
            [
                config_root / "site" / "eps" / "site_add_paths_a.json",
                config_root / "site" / "eps" / "site_add_paths.json",
            ]
        )

        # Check that static and dynamic paths were added in the correct order.
        # Note: `site_add_paths` ends up adding the `site_add_paths_a.json` path
        # twice, the first time the path is encountered, all other instances of
        # that path are discarded.
        assert len(site.paths) == 4
        assert site.paths[0].name == "site_add_paths_c.json"
        assert site.paths[1].name == "site_add_paths_b.json"
        assert site.paths[2].name == "site_add_paths_a.json"
        assert site.paths[3].name == "site_add_paths.json"

        # Check which "set" value was resolved by the end. To correctly process
        # the list returned by the entry_points are processed in reverse order
        assert site["test_data"] == ["site_add_paths_c.json"]

    def test_site_finalize(self, config_root):
        """Test that site entry_point `hab.site.finalize` is called.

        This expects that the entry point will raise a `NotImplementedError` with
        a specific message. This requires that each test has its own site json
        file enabling that specific entry_point. See `tests/site/eps/README.md`.
        """
        with pytest.raises(
            NotImplementedError,
            match="hab_test_entry_points.site_finalize called successfully",
        ):
            Site(
                [
                    config_root / "site" / "eps" / "site_finalize.json",
                ]
            )

    def test_habcache_cls(self, config_root, uncached_resolver):
        """Test that site entry_point `hab.habcache_cls` is respected."""
        # By default `hab.cache.Cache` class is used
        assert isinstance(uncached_resolver.site.cache, Cache)

        # The `hab.habcache_cls` entry_point uses the requested class
        with pytest.raises(
            NotImplementedError,
            match="hab_test_entry_points.CacheVX class was used",
        ):
            Site([config_root / "site" / "eps" / "site_habcache_cls.json"])



import logging
import re
import sys
import traceback
from datetime import datetime
from pathlib import Path

import click
from colorama import Fore

from . import Resolver, Site, __version__, utils
from .parsers.unfrozen_config import UnfrozenConfig

logger = logging.getLogger(__name__)


# No one wants to actually type `--help`
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


class UriArgument(click.Argument):
    """Accepts a URI string, frozen string or path to frozen json file.

    - If a frozen string or json file path is passed returns the unfrozen data
        dictionary.
    - If a uri is provided, the string is returned unmodified.
    - If a `-` is passed and user prefs are enabled, the stored uri in the
        current user's prefs is returned. If a stored uri is not resolved a
        UsageError is raised if required, or returned for later error handling.
        When using a user pref, a message is written to the error stream to
        ensure the user can see what uri was resolved. It is written to the error
        stream so it doesn't interfere with capturing output to a file(json).
    - If the timestamp on the .hab_user_prefs.json is lapse, the user will be
        prompted to address the out of date URI by entering a new path or using
        the already saved path.

    This also handles saving the provided uri to user prefs if enabled by
    `SharedSettings.enable_user_prefs_save`. This is only respected if an uri is
    provided. Ie if a frozen uri, json file or `-` are passed, prefs are not saved.

    Note: Using `err=True` so this output doesn't affect capturing of hab output from
    cmds like `hab dump - --format json > output.json `.
    """

    def __uri_prompt(self, uri=None):
        """Wrapper function of click.prompt.
        Used to get a URI entry from the user"""
        if uri:
            response = click.prompt(
                "Please enter a new URI...\n"
                "Or press ENTER to reuse the expired URI:"
                f" [{Fore.LIGHTBLUE_EX}{uri}{Fore.RESET}]",
                default=uri,
                show_default=False,
                type=str,
                err=True,
            )
        else:
            response = click.prompt(
                "No URI exists.\nPlease Enter a URI", type=str, err=True
            )
        return response

    def type_cast_value(self, ctx, value):
        """Convert and validate the uri value. This override handles saving the
        uri to user prefs if enabled by the cli.
        """
        if value is None:
            result = click.UsageError("Missing argument 'URI'")
            if self.required:
                raise result
            return result
        # User wants to use saved user prefs for the uri
        if value == "-":
            uri_check = ctx.obj.resolver.user_prefs().uri_check()
            # This will indicate that no user_pref.json was saved
            # and the user will be required to enter a uri path.
            if uri_check.uri is None:
                value = self.__uri_prompt()
                # Save a new .hab_user_prefs.json
                ctx.obj.resolver.user_prefs().uri = value
                click.echo("Saved hab user prefs", err=True)
                return value
            # Check if the saved .hab_user_prefs.json has an expire timestamp
            elif uri_check.timedout:
                logger.info(
                    f"{Fore.RED}Invalid 'URI' preference: {Fore.RESET}"
                    f"The saved URI {Fore.LIGHTBLUE_EX}{uri_check.uri}{Fore.RESET} "
                    f"has expired.",
                )
                # The uri is expired so lets ask the user for a new uri
                value = self.__uri_prompt(uri_check.uri)
                if value:
                    # Saving an updated .hab_user_prefs.json
                    ctx.obj.resolver.user_prefs().uri = value
                    click.echo("Saving hab user prefs", err=True)
                    return value
                else:
                    if self.required:
                        raise click.UsageError("A URI is required for Hab use.")
            # user_pref.json is found and its saved uri will be used
            else:
                click.echo(
                    f"Using {Fore.LIGHTBLUE_EX}{uri_check.uri}{Fore.RESET} "
                    "from hab user prefs",
                    err=True,
                )
                # Don't allow users to re-save the user prefs value when using
                # a user prefs value so they don't constantly reset the timeout.
                return uri_check.uri
        # User passed a frozen hab string
        if re.match(r"^v\d+:", value):
            return utils.decode_freeze(value)

        # If its not a string, convert to a Path object, if the path exists,
        # return the extracted dictionary assuming it was a json file.
        try:
            cpath = click.Path(path_type=Path, file_okay=True, resolve_path=True)
            data = cpath.convert(value, None, ctx=ctx)
            if data.exists():
                return utils.json.load(data.open())
        except ValueError:
            self.fail(f"{value!r} is not a valid frozen file path.", ctx)

        # Use standard click type casting on value.
        value = super().type_cast_value(ctx, value)

        # Save the uri to user prefs if requested. This should only happen if the
        # user actually provided a uri, not for other valid inputs on this class.
        if ctx.obj.enable_user_prefs_save:
            ctx.obj.resolver.user_prefs().uri = value

        return value


class UriHelpClass(click.Command):
    """Adds info about user prefs to the help text of commands that take a uri."""

    # Note: the leading whitespace is important to ensure the help text formats
    # correctly when merged with multi-line docstrings like activate.
    uri_text = """
    If you pass a dash `-` for URI, it will use the last URI you saved. You can
    update the saved uri by adding `--save-prefs` to a hab call. For example:
    `hab --save-prefs {subcmd} a/uri`.
    """
    timeout_text = (
        "The saved uri will periodically timeout requiring you to re-save your uri."
    )

    def get_help(self, ctx):
        prefs = ctx.obj.resolver.user_prefs()
        if prefs.enabled:
            self.help += f"\n\n{self.uri_text}".format(subcmd=self.name)
            # Only show timeout info if timeout is enabled
            if prefs.uri_timeout:
                self.help += self.timeout_text
        return super().get_help(ctx)


class SharedSettings(object):
    def __init__(
        self,
        site_paths=None,
        verbosity=0,
        script_dir=None,
        script_ext=None,
        prereleases=None,
        forced_requirements=None,
        dump_scripts=False,
        enable_user_prefs=None,
        enable_user_prefs_save=False,
        cached=True,
    ):
        self.verbosity = verbosity
        self.script_dir = Path(script_dir or ".").resolve()
        self.script_ext = script_ext
        self._resolver = None
        self.prereleases = prereleases
        self.forced_requirements = forced_requirements
        self.dump_scripts = dump_scripts
        self.site_paths = site_paths if site_paths else []
        self.enable_user_prefs = enable_user_prefs
        self.enable_user_prefs_save = enable_user_prefs_save
        self.cached = cached

    @classmethod
    def log_context(cls, uri):
        """Writes a logger.info call for the given uri string or dictionary."""
        if isinstance(uri, dict):
            logger.info("Context: {}".format(uri["uri"]))
        else:
            logger.info("Context: {}".format(uri))

    @property
    def resolver(self):
        if self._resolver is None:
            site = Site(self.site_paths)
            site.cache.enabled = self.cached
            self._resolver = Resolver(
                site=site,
                prereleases=self.prereleases,
                forced_requirements=self.forced_requirements,
            )
            self._resolver.dump_scripts = self.dump_scripts
            self._resolver.user_prefs().enabled = self.enable_user_prefs
        return self._resolver

    def write_script(
        self,
        uri,
        launch=None,
        exit=False,
        args=None,
        create_launch=False,
    ):
        """Generate the script the calling shell scripts expect to setup the environment"""
        self.log_context(uri)
        logger.debug(f"Script dir: {self.script_dir} ext: {self.script_ext}")

        if args:
            # convert to list, subprocess.list2cmdline does not like tuples
            args = list(args)

        if isinstance(uri, dict):
            # Load frozen json data instead of processing the URI
            ret = UnfrozenConfig(uri, self.resolver)
        elif uri is None:
            # If a uri wasn't provided by the user raise the exception click
            # would have raised if uri didn't have `required=False`.
            raise click.UsageError("Missing argument 'URI'.")
        else:
            # Otherwise just process the uri like normal
            ret = self.resolver.resolve(uri)

        ret.write_script(
            self.script_dir,
            self.script_ext,
            launch=launch,
            exit=exit,
            args=args,
            create_launch=create_launch,
        )

    @classmethod
    def set_ctx_instance(cls, ctx, key, value):
        """Set a property on the SharedSetting instance for the given ctx,
        creating the instance if required.

        Args:
            ctx (click.Context): The context passed by click. Sets the requested
                key/value on ctx.obj. If obj isn't set it will create an instance.
            key (str or click.Option): The name of the SharedSettings property to modify.
            value (str): The value to store on key.
        """
        if isinstance(key, click.Option):
            key = key.name

        # Create the instance if it wasn't already created
        if ctx.obj is None:
            ctx.obj = cls()
        elif not isinstance(ctx.obj, SharedSettings):
            raise RuntimeError("Ctx.obj already set to an incompatible class.")

        setattr(ctx.obj, key, value)

        # If the resolver was already created, destroy it to ensure it gets
        # created with the updated options.
        if ctx.obj._resolver:
            # Let devs know if this gets called, ideally it never will be.
            logger.warning("[Optimization warning]: Resetting ctx resolver.")
            ctx.obj._resolver = None

        # Ensure logging settings are properly respected
        if key == "verbosity":
            global _verbose_errors

            if value > 2:
                value = 2
            _verbose_errors = bool(value)

            level = [logging.WARNING, logging.INFO, logging.DEBUG][value]
            logging.basicConfig(level=level)
        elif key == "logging_config":
            if value is not None:
                value = Path(value)
            utils.Platform.configure_logging(value)

        return value

    @property
    def site_paths(self):
        """A list of site json files used to define the Site as Path's."""
        return self._site_paths

    @site_paths.setter
    def site_paths(self, value):
        # Ensure the paths are converted to Path objects.
        self._site_paths = [Path(p) for p in value]


class SiteCommandLoader(click.Group):
    """Allows extending the hab cli by modifying the site config. You can add
    additional `click.Command` objects to the hab cli.

    For example if you add this to your site json file:
    `{"append": {"entry_points": {"hab.cli": {"gui": "hab_gui.cli:gui"}}}}`

    The "gui" key in the innermost dict is used for site resolution so another
    site json file can replace a upper level definition. As a general rule, the
    name should match the exposed click.command function.

    "hab_gui.cli:gui" defines what code to execute. For details on defining this,
    see value for `importlib-metadata.EntryPoints`. In practice this results in
    `from  hab_gui.cli import gui`.
    For the `hab.cli` entry points, its expected that the linked function(`gui`)
    is a `click.Command` object.
    """

    def _cli_entry_points(self, site):
        try:
            return self._ep_cache
        except AttributeError:
            # First call of this method, create the cache
            self._ep_cache = []

        # And populate the cache
        for ep in site.entry_points_for_group("hab.cli"):
            func = ep.load()
            self._ep_cache.append((ep, func))

        return self._ep_cache

    def list_commands(self, ctx):
        ret = super().list_commands(ctx)
        # Add any site defined entry_point commands
        for _, cmd in self._cli_entry_points(ctx.obj.resolver.site):
            ret.append(cmd.name)
        return sorted(ret)

    def get_command(self, ctx, name):
        # Find and use site defined entry_points commands
        for _, funct in self._cli_entry_points(ctx.obj.resolver.site):
            if name == funct.name:
                return funct
        # falling back to the commands defined by hab
        return super().get_command(ctx, name)


# This variable is used to keep track of the user enabled verbose hab output
# using `hab -v ...`. If they do, the full traceback is printed to the shell,
# otherwise just the last part of the traceback is printed so its easier to
# report and hopefully more descriptive.
_verbose_errors = False


# Establish CLI command group
@click.group(cls=SiteCommandLoader, context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__, prog_name="hab")
@click.option(
    "--site",
    "site_paths",
    callback=SharedSettings.set_ctx_instance,
    # Note: Eager enables processing of site when generating help so it can
    # include commands defined by plugins.
    is_eager=True,
    multiple=True,
    # TODO: Implement a custom type that handles pathsep and respects env vars
    type=click.Path(file_okay=True, resolve_path=True),
    help="One or more site json files to load settings from. Uses the env var "
    "`HAB_PATHS` if not passed. The values in each file are merged into a single "
    "dict with the right most value of any given configuration option being used.",
)
@click.option(
    "-v",
    "--verbose",
    "verbosity",
    callback=SharedSettings.set_ctx_instance,
    count=True,
    # Note: Using eager makes it so logging is configured as early as possible
    # based on this argument.
    is_eager=True,
    help="Increase the verbosity of the output. Can be used up to 3 times. "
    "This also enables showing a full traceback if an exception is raised.",
)
@click.option(
    "--logging-config",
    callback=SharedSettings.set_ctx_instance,
    # Note: Using eager makes it so logging is configured as early as possible
    # based on this argument.
    is_eager=True,
    type=click.Path(file_okay=True, resolve_path=True),
    help="Path to json file defining a logging configuration based on "
    "logging.config.dictConfig. If not specified uses .hab_logging_prefs.json "
    "next to user prefs file if it exists.",
)
@click.option(
    "--script-dir",
    callback=SharedSettings.set_ctx_instance,
    type=click.Path(file_okay=False, resolve_path=False),
    help="This directory will contain the shell specific script files to enable "
    "this environment configuration.",
)
@click.option(
    "--script-ext",
    callback=SharedSettings.set_ctx_instance,
    help="The shell specific scripts created in script-dir will have this "
    "format and extension.",
)
@click.option(
    "--cache/--no-cache",
    "cached",
    callback=SharedSettings.set_ctx_instance,
    default=True,
    help="Allow per-site caching of configs and distros.",
)
@click.option(
    "--pre/--no-pre",
    "prereleases",
    callback=SharedSettings.set_ctx_instance,
    default=None,
    help="Include pre-releases when finding the latest distro version.",
)
@click.option(
    "-r",
    "--requirement",
    "forced_requirements",
    callback=SharedSettings.set_ctx_instance,
    multiple=True,
    help="Forces this distro requirement ignoring normally resolved requirements. Using "
    "this may lead to configuring your environment incorrectly, use with caution.",
)
@click.option(
    "--dump-scripts/--no-dump-scripts",
    callback=SharedSettings.set_ctx_instance,
    default=False,
    help=(
        "Print the generated scripts hab uses for this command instead of "
        "running them. This does not work for dump."
    ),
)
@click.option(
    "--prefs/--no-prefs",
    "enable_user_prefs",
    callback=SharedSettings.set_ctx_instance,
    default=None,
    help="If you don't pass a URI, allow looking it up from user prefs.",
)
@click.option(
    "--save-prefs/--no-save-prefs",
    "enable_user_prefs_save",
    callback=SharedSettings.set_ctx_instance,
    default=None,
    help="Update the uri stored in prefs if the uri is provided.",
)
@click.pass_context
def _cli(ctx, **kargs):
    # Note: Using the `set_ctx_instance` callback on the options prevents
    # the need to process anything inside of this function.
    pass


# set uri command
@_cli.command()
@click.argument("uri", required=False)
@click.pass_obj
def set_uri(settings, uri):
    """Allows for saving a local URI default by passing
    a URI argument.  If no argument is passed uri-set
    will prompt you to enter and argument."""
    settings.log_context(uri)
    current_uri = settings.resolver.user_prefs().uri
    if uri is None:
        uri = click.prompt(
            "Please enter a URI value"
            f"[{Fore.LIGHTBLUE_EX}{current_uri}{Fore.RESET}]",
            default=current_uri,
            show_default=False,
            type=str,
            err=True,
        )
    click.echo(f"\nSetting default URI to: {Fore.LIGHTBLUE_EX}{uri}{Fore.RESET}\n")
    settings.resolver.user_prefs().uri = uri


# env command
@_cli.command(cls=UriHelpClass)
@click.argument("uri", cls=UriArgument)
@click.option(
    "-l",
    "--launch",
    default=None,
    help="Run this alias after activating. This leaves the new shell active.",
)
@click.pass_obj
def env(settings, uri, launch):
    """Configures and launches a new shell with the resolved setup."""
    settings.write_script(uri, create_launch=True, launch=launch)


# dump command
@_cli.command(cls=UriHelpClass)
# For specific report_types uri is not required. This is manually checked in
# the code below where it raises `uri_error`.
@click.argument("uri", required=False, cls=UriArgument)
@click.pass_obj
@click.option(
    "--env/--no-env",
    default=True,
    help="Show the environment variable as a flattened structure.",
)
@click.option(
    "--env-config/--no-env-config",
    "--envc/--no-envc",
    default=False,
    help="Show the environment variable as a flattened structure.",
)
@click.option(
    "-t",
    "--type",
    "report_type",
    type=click.Choice(
        ["nice", "site", "s", "uris", "u", "versions", "v", "forest", "f"]
    ),
    default="nice",
    help="Type of report.",
)
@click.option(
    "--flat/--no-flat",
    default=True,
    help="Flatten the resolved object",
)
@click.option(
    "-v",
    "--verbose",
    "verbosity",
    count=True,
    help="Show increasingly detailed output. Can be used up to 3 times.",
)
@click.option(
    "-f",
    "--format",
    "format_type",
    type=click.Choice(["nice", "freeze", "json", "versions"]),
    default="nice",
    help="Choose how the output is formatted.",
)
def dump(settings, uri, env, env_config, report_type, flat, verbosity, format_type):
    """Resolves and prints the requested setup."""

    # Convert uri argument to handle if a uri was not provided/loaded
    uri_error = None
    if isinstance(uri, click.UsageError):
        uri_error = uri
        uri = None

    settings.log_context(uri)
    # Convert report_tupe short names to long names for ease of processing
    report_map = {"u": "uris", "v": "versions", "f": "forest", "s": "site"}
    report_type = report_map.get(report_type, report_type)

    resolver = settings.resolver

    if report_type in ("uris", "versions", "forest"):
        # Allow the user to disable truncation of versions with verbosity flag
        truncate = None if verbosity else 3

        def echo_line(line):
            if line.strip() == line:
                click.echo(f"{Fore.GREEN}{line}{Fore.RESET}")
            else:
                click.echo(line)

        if report_type in ("uris", "forest"):
            click.echo(f'{Fore.YELLOW}{" URIs ".center(50, "-")}{Fore.RESET}')
            # Filter out any URI's hidden by the requested verbosity level
            with utils.verbosity_filter(resolver, verbosity):
                for line in resolver.dump_forest(resolver.configs):
                    echo_line(line)
        if report_type in ("versions", "forest"):
            click.echo(f'{Fore.YELLOW}{" Versions ".center(50, "-")}{Fore.RESET}')
            for line in resolver.dump_forest(
                resolver.distros,
                attr="name",
                truncate=truncate,
            ):
                echo_line(line)
    elif report_type == "site":
        click.echo(resolver.site.dump(verbosity=verbosity))
    else:
        if isinstance(uri, dict):
            # Load frozen json data instead of processing the URI
            ret = UnfrozenConfig(uri, resolver)
        elif uri_error:
            # If the user didn't choose a report type that doesn't require a uri
            # and a uri wasn't specified, raise a UsageError  similar to the one
            # click normally raises. UriArgument likely has augmented the
            # exception to make it easier to debug the issue.
            raise uri_error
        elif flat:
            ret = resolver.resolve(uri)
        else:
            ret = resolver.closest_config(uri)

        # This is a seperate set of if/elif/else statements than from above.
        # I became confused while reading so decided to add this reminder.
        if format_type == "freeze":
            ret = utils.encode_freeze(ret.freeze(), site=resolver.site)
        elif format_type == "json":
            ret = utils.dumps_json(ret.freeze(), indent=2)
        elif format_type == "versions":
            ret = "\n".join([v.name for v in ret.versions])
        else:
            ret = ret.dump(
                environment=env, environment_config=env_config, verbosity=verbosity
            )

        click.echo(ret)


# activate command
@_cli.command(cls=UriHelpClass)
@click.argument("uri", cls=UriArgument)
@click.option(
    "-l",
    "--launch",
    default=None,
    help="Run this alias after activating. This leaves the new shell activated.",
)
@click.pass_obj
def activate(settings, uri, launch):
    """Resolves the setup and updates in the current shell.

    In powershell and bash you must use the source dot: ". hab activate ..."
    WARNING: Not supported in Command Prompt currently.
    """
    if settings.script_ext in (".bat", ".cmd"):
        # The hab shell scripts clean up their temp script files before exiting.
        # Due to batch not having a `function` feature we have to create the
        # aliases as extra script files to support complex aliases. We can't use
        # doskey if we want to support complex aliases. Another reason to avoid
        # doskey, is that it leaks assignments to the parent command prompt when
        # using it for env/launch so exiting doesn't get rid of aliases in those
        # modes. The env/launch commands continue to run till the user exits, but
        # activate exits immediately so the temp alias batch files get deleted
        # before they can be used by the user/script breaking this command.
        click.echo(
            f"{Fore.RED}Not Supported:{Fore.RESET} Using hab activate in the "
            "Command Prompt is not currently supported."
        )
        sys.exit(1)

    settings.write_script(uri, launch=launch)


# launch command
@_cli.command(
    context_settings=dict(
        ignore_unknown_options=True,
    ),
    cls=UriHelpClass,
)
@click.argument("uri", cls=UriArgument)
@click.argument("alias")
# Pass all remaining arguments to the requested alias
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def launch(settings, uri, alias, args):
    """Configure and launch an alias without modifying the current shell. The
    first argument is a URI, The second argument is the ALIAS to launch. Any
    additional arguments are passed as launch arguments to the alias. Note, if
    using bash on windows you may need to pass file paths correctly for bash as
    any quotes used may not make it to the alias launch arguments. (ie: '/c/Program\\ Files').
    """
    settings.write_script(uri, create_launch=True, launch=alias, exit=True, args=args)


# Cache command
@_cli.command()
@click.argument("path", type=click.Path(file_okay=True, resolve_path=True))
@click.pass_obj
def cache(settings, path):
    """Create/update the cache for a given site file. The path argument is the
    site config file. To allow for cross-platform support you should make sure
    you are loading the same site configuration that will be used by this cache
    or at least a site configuration that defines the same `platform_path_maps`.
    """
    path = Path(path)
    click.echo(f"Caching: {path}")
    s = datetime.now()
    out = settings.resolver.site.cache.save_cache(settings.resolver, path)
    e = datetime.now()
    click.echo(f"Cache took: {e - s}, cache file: {out}")


def cli(*args, **kwargs):
    """Runs the hab cli. If an exception is raised, only the exception message
    is printed and the stack trace is hidden. Use `hab -v ...` to enable showing
    the entire stack trace.
    """
    try:
        return _cli(*args, **kwargs)
    except Exception:
        click.echo(f"{Fore.RED}Hab encountered an error:{Fore.RESET}")
        if _verbose_errors:
            # In verbose mode show the full traceback
            raise

        # Print the traceback message without the stack trace.
        click.echo(traceback.format_exc(limit=0))
        return 1


if __name__ == "__main__":
    cli()



import os
import string

from . import utils


class Formatter(string.Formatter):
    """A extended string formatter class to parse habitat configurations.

    Adds support for the "!e" conversion flag. This will fill in the key as a properly
    formatted environment variable specifier. For example ''{PATH!e}'' will be converted
    to ``$PATH`` for the sh language, and ``%env:PATH`` for the ps language. You can
    convert "!e" to "!s" by setting expand to True. This simulates `os.path.expandvars`.

    This also converts ``{;}`` to the language specific path separator for environment
    variables. On linux this is ``:`` on windows(even in bash) this is ``;``.

    You need to specify the desired language when initializing this class. This can be
    one of the supported language keys in ''Formatter.shell_formats'', or you can pass
    a file extension supported by `Formatter.language_from_ext`. If you pass None, it
    will preserve the formatting markers so it can be converted later.
    """

    shell_formats = {
        # Command Prompt
        "batch": {
            "env_var": "%{}%",
            ";": ";",
        },
        # Power Shell
        "ps": {
            "env_var": "$env:{}",
            ";": ";",
        },
        # Using bash on linux
        "sh": {
            "env_var": "${}",
            ";": ":",
        },
        # Using bash on windows
        "shwin": {
            "env_var": "${}",
            ";": ":",
        },
        # Delay the format for future calls. This allows us to process the environment
        # without changing these, and then when write_script is called the target
        # scripting language is resolved.
        None: {
            "env_var": "{{{}!e}}",
            ";": "{;}",
        },
    }

    def __init__(self, language, expand=False):
        super().__init__()
        self.language = self.language_from_ext(language)
        self.current_field_name = None
        self.expand = expand

    def convert_field(self, value, conversion):
        if conversion == "e":
            # Expand the env var to the real string value simulating `os.path.expandvars`
            if self.expand:
                return super().convert_field(value, "s")

            # Otherwise insert the correct shell script env var reference
            return self.shell_formats[self.language]["env_var"].format(
                self.current_field_name
            )

        return super().convert_field(value, conversion)

    def get_field(self, field_name, args, kwargs):
        self.current_field_name = field_name
        ret = super().get_field(field_name, args, kwargs)
        return ret

    @classmethod
    def language_from_ext(cls, ext):
        """Turn the provided file ext into the common name for that language.
        This can be used to find the correct data for the cls.shell_formats mapping.

        ".bat" and ".cmd" return "batch". ".ps1" returns "ps". ".sh" or an empty
        string return "sh" if on linux and "shwin" if on windows. If `None` is passed
        the format will be replaced with the same format command so future format
        calls can re-apply the changes. Any other value passed is returned unmodified.
        """
        if ext in (".bat", ".cmd"):
            return "batch"
        elif ext == ".ps1":
            return "ps"
        elif ext in (".sh", ""):
            # Assume no ext is a .sh file
            if utils.Platform.name() == "windows":
                return "shwin"
            else:
                return "sh"
        return ext

    def merge_kwargs(self, kwargs):
        """Merge the provided kwargs on top of the current language's shell_formats
        dict. This makes it so the default format options are added by default, but
        still allows us to override them if required
        """
        ret = dict(self.shell_formats[self.language], **kwargs)
        ret = dict(os.environ, **ret)
        return ret

    def vformat(self, format_string, args, kwargs):
        kwargs = self.merge_kwargs(kwargs)
        ret = super().vformat(format_string, args, kwargs)
        return ret



import logging
import os
import subprocess
from pathlib import Path

import anytree
from colorama import Fore, Style
from jinja2 import Environment, FileSystemLoader
from packaging.version import Version

from .. import NotSet, utils
from ..errors import DuplicateJsonError, HabError, ReservedVariableNameError
from ..formatter import Formatter
from ..site import MergeDict
from ..solvers import Solver
from .meta import HabMeta, hab_property

logger = logging.getLogger(__name__)

TEMPLATES = Path(__file__).parent.parent / "templates"


class HabBase(anytree.NodeMixin, metaclass=HabMeta):
    """Base class for the various parser classes. Provides most of the functionality
    to parse a json configuration file and resolve it for use in hab.

    Args:
        forest (dict): A dictionary map used to calculate the context when resolving
            this object.
        resolver (hab.Resolver): The Resolver used to lookup requirements.
        filename (str, optional): Automatically call load on this filename.
        parent (hab.parsers.HabBase, optional): Parent for this object.
        root_paths (set, optional): The base glob path being processed to create the
            HabBase objects for this forest. If two
    """

    # Subclasses can change this to control how data is tweaked by the load method.
    _context_method = "key"
    # A instance of this class is used to build a parent anytree item if no
    # configuration was processed yet to fill in that node. This is set to the
    # Placeholder class after it is defined below. This allows the
    # DistroVersion class to use Distro as its placeholder
    _placeholder = None

    def __init__(self, forest, resolver, filename=None, parent=None, root_paths=None):
        super().__init__()
        self.frozen_data = {}
        self._filename = None
        self._dirname = None
        self._distros = NotSet
        self._variables = NotSet
        self._uri = NotSet
        self.parent = parent
        self.root_paths = set()
        if root_paths:
            self.root_paths = root_paths
        self.forest = forest
        self.resolver = resolver
        if filename:
            self.load(filename)

    def __repr__(self):
        cls = type(self)
        return "{}.{}('{}')".format(cls.__module__, cls.__name__, self.fullpath)

    def _cache(self):
        return {}

    def _collect_values(self, node, props=None, default=False):
        """Recursively process this config node and its parents until all
        missing_values have been resolved or we run out of parents.

        Args:
            node (HabBase): This node's values are copied to self as long as
                they are not NotSet.
            props (list, optional): The props to process, if None is
                passed then uses `hab_property` values respecting sort_key.
            default (bool, optional): Enables processing the default nodes as
                part of this methods recursion. Used for internal tracking.
        """
        logger.debug("Loading node: {} inherits: {}".format(node.name, node.inherits))
        if props is None:
            props = sorted(
                self._properties, key=lambda i: self._properties[i].sort_key()
            )

        self._missing_values = False
        # Use sort_key to ensure the props are processed in the correct order
        for attrname in props:
            if getattr(self, attrname) != NotSet:
                continue
            if attrname == "alias_mods":
                if hasattr(node, "alias_mods") and node.alias_mods:
                    self._alias_mods = {}
                    # Format the alias environment at this point so any path
                    # based variables like {relative_root} are resolved against
                    # the node's directory not the alias being modified
                    mods = node.format_environment_value(node.alias_mods)
                    for name, mod in mods.items():
                        self._alias_mods.setdefault(name, []).append(mod)
                continue
            value = getattr(node, attrname)
            if value is NotSet:
                self._missing_values = True
            else:
                setattr(self, attrname, value)

        if node.inherits and self._missing_values:
            parent = node.parent
            if parent:
                return self._collect_values(parent, props=props, default=default)
            elif not default and "default" in self.forest:
                # Start processing the default setup
                default = True
                default_node = self.resolver.closest_config(node.fullpath, default=True)
                self._collect_values(default_node, props=props, default=default)

        return self._missing_values

    @classmethod
    def _dump_versions(cls, value, verbosity=0, color=None):
        """Returns the version information for this object as a list of strings."""
        if verbosity < 3:
            # Only show the names of the versions
            return sorted([v.name for v in value], key=lambda i: i.lower())

        # Include the definition of the version's path for debugging
        if color:  # pragma: no cover
            fmt = f"{Fore.GREEN}{{}}{Style.RESET_ALL}:  {{}}"
        else:
            fmt = "{}:  {}"
        return [
            fmt.format(v.name, v.filename)
            for v in sorted(value, key=lambda i: i.name.lower())
        ]

    def check_environment(self, environment_config):
        """Check that the environment config only makes valid adjustments.
        For example, check that we never replace path, only append/prepend are allowed.
        """
        for operation in ("unset", "set", "prepend", "append"):
            if operation not in environment_config:
                continue

            keys = environment_config.get(operation, [])
            if operation == "set":
                # set is a dictionary while unset is a list
                keys = keys.keys()

            for key in keys:
                key_upper = key.upper()

                # These are reserved environment variables that can not be configured
                # by configs and distros.
                if key_upper == "HAB_URI":
                    raise KeyError(f'"{key_upper}" is a reserved environment variable')

                # We can't clear "PATH" or it would likely break the shell and
                # application execution.
                if key_upper == "PATH":
                    msg = None
                    if operation == "set":
                        key = environment_config[operation][key]
                        msg = 'You can not use PATH for the set operation: "{}"'
                    elif operation == "unset":
                        msg = "You can not unset PATH"
                    if msg:
                        raise ValueError(msg.format(key))

    def check_min_verbosity(self, config):
        """Return if the given config should be visible based on the resolver's
        current verbosity settings. Returns True if there isn't a resolver.
        """
        if not self.resolver:
            return True

        # Respect `hab.utils.verbosity_filter` with context settings
        target = self.resolver._verbosity_target
        current = self.resolver._verbosity_value

        if current is None:
            # If None, always show all results
            return True

        min_verbosity = self.get_min_verbosity(config, target)
        return current >= min_verbosity

    @property
    def context(self):
        """A list of parent context strings.

        The resolved URI parents of the this object. This does not include the name
        of the object, only its parents. `project_a/Sc001` would resolve into context:
        `["project_a"]` and name: `"Sc001"`.
        """
        return self.frozen_data.get("context", NotSet)

    @context.setter
    def context(self, context):
        self.frozen_data["context"] = context

        if not self.context:
            # Add the root of this tree to the forest
            if self.name in self.forest:
                if not isinstance(self.forest[self.name], self._placeholder):
                    msg = 'Can not add "{}", the context "{}" it is already set'.format(
                        self.filename,
                        self.fullpath,
                    )
                    if self.forest[self.name].root_paths.intersection(self.root_paths):
                        # If one of the root_paths was already added to target, then
                        # the same context was defined inside a folder structure.
                        # We  only support this from unique config/distro paths like if
                        # a developer is overriding a new version of a config.
                        raise DuplicateJsonError(msg)
                    else:
                        logger.warning(msg)
                        # Document that we have processed the new paths so if there are
                        # any duplicate context:name defined in these paths, the above
                        # exception is raised.
                        self.forest[self.name].root_paths.update(self.root_paths)

                        # Do not run the rest of this method
                        return

                # Preserve the children of the placeholder object if it exists
                self.children = self.forest[self.name].children
            self.forest[self.name] = self
            logger.debug("Add to forest: {}".format(self))
        else:
            resolver = anytree.Resolver("name")
            # Get the tree root
            root_name = self.context[0]
            if root_name in self.forest:
                root = self.forest[root_name]
                logger.debug("Using root: {}".format(root.fullpath))
            else:
                root = self._placeholder(self.forest, self.resolver)
                root.name = root_name
                self.forest[root_name] = root
                logger.debug("Created placeholder root: {}".format(root.fullpath))

            # Process the intermediate parents
            for child_name in self.context[1:]:
                try:
                    root = resolver.get(root, child_name)
                    logger.debug("Found intermediary: {}".format(root.fullpath))
                except anytree.resolver.ResolverError:
                    root = self._placeholder(self.forest, self.resolver, parent=root)
                    root.name = child_name
                    logger.debug(
                        "Created placeholder intermediary: {}".format(root.fullpath)
                    )

            # Add this node to the tree
            try:
                target = resolver.get(root, self.name)
            except anytree.resolver.ResolverError:
                # There is no placeholder, just add self as a child
                self.parent = root
                logger.debug("Adding to parent: {}".format(root.fullpath))
            else:
                if isinstance(target, self._placeholder) and target.name == self.name:
                    # replace the placeholder with self
                    self.parent = target.parent
                    self.children = target.children
                    # Remove the placeholder from the tree
                    target.parent = None
                    logger.debug("Removing placeholder: {}".format(target.fullpath))
                else:
                    msg = 'Can not add "{}", the context "{}" it is already set'.format(
                        self.filename,
                        self.fullpath,
                    )
                    if target.root_paths.intersection(self.root_paths):
                        # If one of the root_paths was already added to target, then
                        # the same context was defined inside a folder structure.
                        # We  only support this from unique config/distro paths like if
                        # a developer is overriding a new version of a config.
                        raise DuplicateJsonError(msg)
                    else:
                        logger.warning(msg)

                        # Document that we have processed the new paths so if there are
                        # any duplicate context:name defined in these paths, the above
                        # exception is raised.
                        target.root_paths.update(self.root_paths)
                        # Do not run the rest of this method
                        return

    @property
    def dirname(self):
        """The directory name of `self.filename`. This value is used to by
        `self.format_environment_value` to fill the "relative_root" variable.
        """
        return self._dirname

    # Note: 'distros' needs to be processed before 'environment'
    @hab_property(verbosity=3, process_order=50)
    def distros(self):
        """A list of all of the requested distros to resolve."""
        return self._distros

    @distros.setter
    def distros(self, distros):
        # Ensure the contents are converted to Requirement objects
        if distros:
            distros = Solver.simplify_requirements(distros)
        self._distros = distros

    def dump(self, environment=True, environment_config=False, verbosity=0, color=None):
        """Return a string of the properties and their values.

        Args:
            environment (bool, optional): Show the environment value.
            environment_config (bool, optional): Show the environment_config value.
            verbosity (int, optional): More information is shown with higher values.
            color (bool, optional): Add console colorization to output. If None,
                respect the site property "colorize" defaulting to True.

        Returns:
            str: The configuration converted to a string
        """
        if color is None:  # pragma: no cover
            color = self.resolver.site.get("colorize", True)

        ret = []
        # Update what properties are shown in the dump
        props = self._properties.copy()
        for k, v in (
            ("environment", environment),
            ("environment_config", environment_config),
        ):
            if not v:
                del props[k]

        for prop in sorted(props, key=lambda p: (props[p].group, p)):
            # Ignore any props with higher verbosity than requested
            pv = props[prop].verbosity
            if pv is None or verbosity < pv:
                # Setting the property's verbosity to None should hide the property
                continue

            flat_list = False
            if prop == "aliases":
                # Filter out any aliases hidden by the requested verbosity level
                with utils.verbosity_filter(self.resolver, verbosity):
                    value = self.aliases
            else:
                value = getattr(self, prop)
            if prop == "aliases" and not value:
                # Don't show the aliases row if none are set
                continue
            elif prop == "aliases" and verbosity < 3:
                value = value.keys()
                flat_list = True
            elif prop == "versions":
                value = self._dump_versions(value, verbosity=verbosity, color=color)

            ret.append(
                utils.dump_object(
                    value, label=f"{prop}:  ", flat_list=flat_list, color=color
                )
            )

        ret = "\n".join(ret)
        name = type(self).__name__
        if color:  # pragma: no cover
            # Only colorize the uri name not the entire title line
            title = (
                f"Dump of {name}({Fore.GREEN}'" f"{self.fullpath}'{Style.RESET_ALL})"
            )
        else:
            title = f"Dump of {name}('{self.fullpath}')"
        return utils.dump_title(title, ret, color=False)

    # Note: 'distros' needs to be processed before 'environment'
    @hab_property(verbosity=2, process_order=80)
    def environment(self):
        """A resolved set of environment variables that should be applied to
        configure an environment. Any values containing a empty string indicate
        that the variable should be unset.
        """
        if "environment" in self.frozen_data:
            return self.frozen_data["environment"].get(utils.Platform.name(), {})

        self.frozen_data["environment"] = {}

        # Merge the environment_config if defined
        if self.environment_config is not NotSet:
            self.merge_environment(self.environment_config)

        return self.frozen_data["environment"].get(utils.Platform.name(), {})

    @hab_property(verbosity=2)
    def environment_config(self):
        """A dictionary of operations to perform on environment variables.

        The top level dictionary defines the modification operation. Valid keys are
        "set", "unset", "prepend" and "append". Each of these keys contain the
        environment variable key/value pairs to be modified.

        Like Rez, the first set, prepend or append operation on a variable will replace
        the existing variable value. This quote from the Rez documentation explains why:
        "Why does this happen? Consider PYTHONPATH - if an initial overwrite did not
        happen, then any modules visible on PYTHONPATH before the rez environment was
        configured would still be there. This would mean you may not have a properly
        configured environment. If your system PyQt were on PYTHONPATH for example,
        and you used rez-env to set a different PyQt version, an attempt to import
        it within the configured environment would still, incorrectly, import the
        system version."

        Note however that this does not apply to the PATH env variable. You may
        only use append and prepend on that variable.
        """
        return self.frozen_data.get("environment_config", NotSet)

    @environment_config.setter
    def environment_config(self, env):
        self.frozen_data["environment_config"] = env
        self.frozen_data.pop("environment", None)

    @hab_property(verbosity=3)
    def filename(self):
        """The filename that defined this object. Any relative paths are
        relative to the directory of this file.
        """
        return self._filename

    @filename.setter
    def filename(self, filename):
        """A path like object. If a empty string is provided, ``Path(os.devnull)``
        will be used to prevent using an relative path.
        """
        # Cache the dirname so we only need to look it up once
        if not filename:
            self._filename = Path(os.devnull)
            self._dirname = Path(os.devnull)
        else:
            self._filename = Path(filename)
            self._dirname = self._filename.parent

    def format_environment_value(self, value, ext=None, platform=None):
        """Apply standard formatting to environment variable values.

        Args:
            value (str): The string to format
            ext (str, optional): Language passed to ``hab.formatter.Formatter``
                for special formatters. In most cases this should not be used.
            platform (str, optional): Convert path values from the current
                platform to this path using `site.plaform_path`.

        Format Keys:
            relative_root: Add the dirname of self.filename or a empty string. Equivalent
                of using the `.` object for file paths. This removes the
                ambiguity of if a `.` should be treated as a relative file path
                or a literal relative_root. Houdini doesn't support the native slash
                direction on windows, so backslashes are replaced with forward slashes.
        """
        if isinstance(value, list):
            # Format the individual items if a list of args is used.
            return [
                self.format_environment_value(v, ext=ext, platform=platform)
                for v in value
            ]
        elif isinstance(value, dict):
            # Format the values each dictionary pair
            return {
                k: self.format_environment_value(v, ext=ext, platform=platform)
                for k, v in value.items()
            }
        elif isinstance(value, (bool, int)):
            # Just return boolean values, no need to format
            return value

        # Include custom variables in the format dictionary.
        kwargs = {}
        if self.variables:
            kwargs.update(self.variables)

        # Custom variables do not override hab variables, ensure they are set
        # to the correct value.
        kwargs["relative_root"] = utils.path_forward_slash(self.dirname)
        ret = Formatter(ext).format(value, **kwargs)

        # Use site.the platform_path_maps to convert the result to the target platform
        if platform:
            ret = utils.path_forward_slash(
                self.resolver.site.platform_path_map(ret, platform)
            )

        return ret

    @property
    def fullpath(self):
        return self.separator.join([node.name for node in self.path])

    @classmethod
    def get_min_verbosity(cls, config, target, default=0):
        """Gets the desired min_verbosity setting for a given dictionary. If the
        desired target is not specified returns the global target, returning default
        if not defined.

        Args:
            config (dict): The config dict to get the desired verbosity level.
                Returns default if None is passed.
            target (str): The name of the desired verbosity level to return if defined.
            default (int, optional): Returned if unable to find a defined value.
        """
        if not config:
            return default
        # Get the min_verbosity dictionary
        verbosity = config.get("min_verbosity", {})
        if verbosity is NotSet:
            return default
        # If the requested target is defined, return its value
        if target in verbosity:
            return verbosity[target]
        # Otherwise return the default global value, defaulting to zero if
        # global was not defined
        return verbosity.get("global", default)

    @property
    def inherits(self):
        """Should this node inherit from a parent."""
        # Note: Sub-classes need to override this method to enable inheritance.
        return False

    def _load(self, filename, cached=True):
        """Sets self.filename and parses the json file returning the data dict.

        Args:
            filename (pathlib.Path): The file to load.
            cached (bool, optional): Enables loading of cached data instead of
                loading the data from disk.
        """
        self.filename = Path(filename)

        if cached:
            ret = self._cache().get(Path(filename).as_posix())
            if ret:
                logger.debug(f'Cached: "{filename}"')
                return ret

        logger.debug(f'Loading "{filename}"')
        return utils.load_json_file(self.filename)

    def load(self, filename, data=None):
        """Load this objects configuration from the given json filename.

        Args:
            filename (str): The json file to load the config from.
            data (dict, optional): If provided this dict is used instead of parsing
                the json file. In this case filename is ignored.
        """
        if data is None:
            data = self._load(filename)

        # Check for NotSet so sub-classes can set values before calling super
        if self.name is NotSet:
            self.name = data["name"]
        if self.variables is NotSet:
            self.variables = data.get("variables", NotSet)
        if "version" in data and self.version is NotSet:
            self.version = data.get("version")
        if self.distros is NotSet:
            self.distros = data.get("distros", NotSet)
        if self.environment_config is NotSet:
            self.environment_config = data.get("environment", NotSet)
        if self.min_verbosity is NotSet:
            self.min_verbosity = data.get("min_verbosity", NotSet)

        if self.context is NotSet:
            # TODO: make these use override methods
            if self._context_method == "key":
                self.context = data.get("context", NotSet)
            elif self._context_method == "name":
                self.context = [data["name"]]

        return data

    def merge_environment(self, environment_config, obj=None):
        """Check and update environment with the provided environment config."""

        if obj is None:
            obj = self

        if environment_config is NotSet:
            # No environment_config dictionary was set, noting to do
            return

        merger = MergeDict(
            relative_root=self.dirname,
            platforms=self.resolver.site["platforms"],
            site=self.resolver.site,
        )
        merger.formatter = obj.format_environment_value
        merger.validator = self.check_environment
        # Flatten the site configuration down to per-platform configurations
        merger.apply_platform_wildcards(
            environment_config, output=self.frozen_data["environment"]
        )

    @hab_property(verbosity=2)
    def min_verbosity(self):
        return self.frozen_data.get("min_verbosity", NotSet)

    @min_verbosity.setter
    def min_verbosity(self, min_verbosity):
        self.frozen_data["min_verbosity"] = min_verbosity

    @hab_property(verbosity=1, group=0, process_order=40)
    def name(self):
        """The name of this object. See ``.context`` for how this is built into
        a full URI. `project_a/Sc001` would resolve into context: `["project_a"]`
        and name: `"Sc001"`.
        """
        return self.frozen_data.get("name", NotSet)

    @name.setter
    def name(self, name):
        self.frozen_data["name"] = name

    def reduced(self, resolver=None, uri=None):
        """Returns a new instance with the final settings applied respecting inheritance"""
        from . import FlatConfig

        return FlatConfig(self, resolver, uri=uri)

    @classmethod
    def shell_escape(cls, ext, value):
        """Apply any shell specific formatting like escape characters."""
        if ext == ".ps1":
            if isinstance(value, list):
                value = " ".join([v.replace(" ", "` ") for v in value])
            else:
                return value.replace(" ", "` ")
        if ext in (".sh", ""):
            # wrapping items in quotes takes care of escaping file paths
            if isinstance(value, list):
                value = subprocess.list2cmdline(value)
            else:
                return '"{}"'.format(value)
        if ext in (".bat", ".cmd"):
            if isinstance(value, list):
                value = subprocess.list2cmdline(value)
            else:
                return '"{}"'.format(value)
        return value

    @classmethod
    def shell_formats(cls, ext):
        """Returns a file ext specific dict that is used to write launch scripts"""
        ret = {}
        if ext in (".bat", ".cmd"):
            ret["launch"] = 'cmd.exe /k "{path}"\n'
        elif ext == ".ps1":
            ret["launch"] = (
                'powershell.exe{launch_args} -ExecutionPolicy Unrestricted -File "{path}"\n'
                "exit $LASTEXITCODE\n"
            )
        elif ext in (".sh", ""):  # Assume no ext is a .sh file
            # `set -e` propagates the error code returned by bash to the caller
            ret["launch"] = 'set -e\nbash --init-file "{path}"\n'

        return ret

    @property
    def uri(self):
        """The Uniform Resource Identifier for this object. This is a combination of
        its context and name properties separated by `/`"""
        if self._uri:
            return self._uri
        return self.fullpath

    def update_environ(self, env, alias_name=None, include_global=True, formatter=None):
        """Updates the given environment variable dictionary to conform with
        the hab environment specification.

        Args:
            env (dict): This dictionary is modified according to the hab
                definition. Often you will want to pass a copy of `os.environ`.
            alias_name (str, optional): If passed also apply any complex alias
                specific environment variable changes.
            include_global (bool, optional): Used to disable adding the global
                hab managed env vars. Disable this and use alias_name to only
                get the env vars set by the alias, not the global ones. This
                also adds `HAB_FREEZE` if possible.
            formatter (hab.formatter.Formatter, optional): Str formatter class
                used to format the env var values.
        """
        ext = utils.Platform.default_ext()
        if formatter is None:
            # Make sure to expand environment variables when formatting.
            formatter = Formatter(ext, expand=True)

        def _apply(data):
            for key, value in data.items():
                if value:
                    value = utils.Platform.collapse_paths(value, ext=ext, key=key)
                    value = formatter.format(value, key=key, value=value)
                    env[key] = value
                else:
                    env.pop(key, None)

        if include_global:
            _apply(self.environment)
            # Add the HAB_FREEZE environment variable documenting the entire
            # resolved hab configuration.
            if hasattr(self, "freeze") and "HAB_FREEZE" not in env:
                env["HAB_FREEZE"] = utils.encode_freeze(
                    self.freeze(), site=self.resolver.site
                )
        if alias_name:
            _apply(self.aliases[alias_name].get("environment", {}))

    @hab_property(verbosity=3)
    def variables(self):
        """A configurable dict of reusable key/value pairs to use when formatting
        text strings in the rest of the json file. This value is stored in the
        `variables` dictionary of the json file.
        """
        return self._variables

    @variables.setter
    def variables(self, variables):
        # Raise an exception if a reserved variable name is used.
        if variables and isinstance(variables, dict):
            invalid = ReservedVariableNameError._reserved_variable_names.intersection(
                variables
            )
            if invalid:
                raise ReservedVariableNameError(invalid, self.filename)
        self._variables = variables

    @property
    def version(self):
        """A `packaging.version.Version` representing the version of this object."""
        return self.frozen_data.get("version", NotSet)

    @version.setter
    def version(self, version):
        if version and not isinstance(version, Version):
            version = Version(version)
        self.frozen_data["version"] = version

    def generate_config_script(
        self,
        template,
        ext,
        alias_dir=None,
        args=None,
        create_launch=False,
        exit=False,
        launch=None,
    ):
        """Build a shell script for the requested template.

        Args:
            template (str): The name of the Jinja2 template file to generate with.
            ext (str): The file extension of the script to generate including leading `.`.

        Returns:
            str: The rendered script.
        """

        environment = Environment(
            loader=FileSystemLoader(str(TEMPLATES)),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = environment.get_template(f"{template}{ext}")
        kwargs = dict(
            alias_dir=alias_dir,
            args=args,
            create_launch=create_launch,
            exit=exit,
            ext=ext,
            formatter=Formatter(ext),
            freeze=None,
            hab_cfg=self,
            launch_info={},
            utils=utils,
        )
        if hasattr(self, "freeze"):
            kwargs["freeze"] = utils.encode_freeze(
                self.freeze(), site=self.resolver.site
            )
        if launch:
            # Write additional args into the launch command. This may not properly
            # add quotes around windows file paths if writing a shell script.
            # At this point we have lost the original double quote the user used.
            if isinstance(args, list):
                if ext == ".ps1":
                    args = " {}".format(subprocess.list2cmdline(args))
                else:
                    args = " {}".format(self.shell_escape(ext, args))
            else:
                args = ""

            # Get the cmd to launch, raising useful errors if invalid
            if launch not in self.aliases:
                raise HabError(f'"{launch}" is not a valid alias name')
            alias = self.aliases.get(launch, {})

            try:
                cmd = alias["cmd"]
            except KeyError:
                raise HabError(
                    f'Alias "{launch}" does not have "cmd" defined'
                ) from None

            cmd = self.shell_escape(ext, cmd)
            kwargs["launch_info"] = dict(key=launch, value=cmd, args=args)

        # Note: jinja2' seems to be inconsistent with its trailing newlines depending
        # on the template and its if statements, so force a single trailing newline
        return template.render(**kwargs).rstrip() + "\n"

    def generate_alias_script(self, template, ext, alias, cfg):
        """Build a shell script for aliases that require their own script files(batch).

        Args:
            template (str): The name of the Jinja2 template file to generate with.
            ext (str): The file extension of the script to generate including leading `.`.
            alias (str): The name of the alias to generate.
            cfg (dict): The configuration of the alias.

        Returns:
            str: The rendered script.
        """
        environment = Environment(
            loader=FileSystemLoader(str(TEMPLATES)),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = environment.get_template(f"{template}{ext}")
        kwargs = dict(
            alias=alias,
            cfg=cfg,
            ext=ext,
            formatter=Formatter(ext),
            hab_cfg=self,
            utils=utils,
        )

        # Note: jinja2' seems to be inconsistent with its trailing newlines depending
        # on the template and its if statements, so force a single trailing newline
        return template.render(**kwargs).rstrip() + "\n"

    def _write_script(self, script_path, content):
        """Work function to save content to script script_path. If dump_scripts
        is enabled, script_path is ignored and content is printed instead.
        """
        if self.resolver.dump_scripts:
            self.print_script(script_path, content)
        else:
            with script_path.open("w") as fle:
                fle.write(content)

    def write_script(
        self, script_dir, ext, launch=None, exit=False, args=None, create_launch=False
    ):
        """Write the configuration to a script file to be run by terminal."""
        script_dir = Path(script_dir)
        config_script = script_dir / f"hab_config{ext}"
        alias_dir = script_dir / "aliases"
        shell = self.shell_formats(ext)

        content = self.generate_config_script(
            "config",
            ext,
            alias_dir=alias_dir,
            args=args,
            create_launch=create_launch,
            exit=exit,
            launch=launch,
        )
        self._write_script(config_script, content)

        # Handle aliases that can't be defined in memory(batch) by writing
        # additional scripts to disk
        if ext in (".bat", ".cmd") and hasattr(self, "aliases"):
            if not self.resolver.dump_scripts:
                alias_dir.mkdir(exist_ok=True)

            for alias, cfg in self.aliases.items():
                content = self.generate_alias_script("alias", ext, alias=alias, cfg=cfg)
                self._write_script(alias_dir / f"{alias}{ext}", content)

        if create_launch:
            launch_script = script_dir / f"hab_launch{ext}"
            launch_args = ""
            if ext == ".ps1":
                # If we want PowerShell to stay open after the script exits pass
                # `-NoExit` to the shell launch. Other shells allow us to simply
                # call exit at the end of the configure script.
                if not (exit and create_launch):
                    launch_args = " -NoExit"

            content = shell["launch"].format(
                path=config_script, launch_args=launch_args
            )
            self._write_script(launch_script, content)

    def print_script(self, filename, content):
        """Prints a header including the filename and content.

        Args:
            filename (pathlib.Path): The name of the file contents is to be
                written to.
            content (str): Shell script that is to be written to filename.
                If colorize is enabled, then syntax highlighting is applied
                based on filenames extension.
        """
        colorize = self.resolver.site.get("colorize", True)
        if colorize:
            from pygments import highlight
            from pygments.formatters import TerminalFormatter
            from pygments.lexers import get_lexer_for_filename

        # Print the name of the script being printed
        if colorize:
            header = f"{Fore.GREEN}-- Script: {filename} --{Fore.RESET}"
        else:
            header = f"-- Script: {filename} --"
        print(header)

        # Highlight and print the script contents
        if colorize:
            lexer = get_lexer_for_filename(filename)
            content = highlight(content, lexer, TerminalFormatter())

        print(content)



import json

import pytest

from hab import utils
from hab.merge_dict import MergeDict


@pytest.mark.parametrize(
    "filename,platforms",
    (
        # if os_specific is not set, add the variables to all platforms
        ("os_agnostic.json", None),
        # If platform is specified, only the requested platforms are returned
        ("os_agnostic_platforms.json", ("reactos",)),
        # os_specific is set, so merge "*" with each platform
        ("os_specific.json", None),
        # if os_specific is set to False, treat it the same as os_agnostic
        ("os_specific_false.json", None),
        # If platform is specified in an os_specific context, only the requested
        # platforms are returned
        ("os_specific_platforms.json", ("reactos",)),
    ),
)
def test_apply_platform_wildcards(config_root, resolver, filename, platforms):
    """Test merging of "*" os specific dicts into platform specific dictionaries"""
    json_path = config_root / "merge_dict" / filename

    with json_path.open() as fle:
        test_data = json.load(fle)

    # The data to feed the apply_platform_wildcards method
    in_data = test_data["in_data"]
    # The output from the apply_platform_wildcards method to check against
    out_data = test_data["out_data"]

    kwargs = {}
    if platforms is not None:
        kwargs["platforms"] = platforms
        kwargs["site"] = resolver.site

    merger = MergeDict(**kwargs)
    result = merger.apply_platform_wildcards(in_data)

    assert result == out_data


def test_path_split(monkeypatch):
    # Check the pathsep argument is respected
    assert utils.Platform.path_split("a:b", pathsep=":") == ["a", "b"]
    assert utils.Platform.path_split("a;b", pathsep=";") == ["a", "b"]
    assert utils.Platform.path_split("a-b", pathsep="-") == ["a", "b"]

    # Check if pathsep argument is not passed, the current os is respected
    # Windows
    monkeypatch.setattr(utils, "Platform", utils.WinPlatform)
    assert utils.Platform.path_split("a;b") == ["a", "b"]
    assert utils.Platform.path_split("a:b") == ["a:b"]
    # Linux/OSX
    monkeypatch.setattr(utils, "Platform", utils.LinuxPlatform)
    assert utils.Platform.path_split("a;b") == ["a;b"]
    assert utils.Platform.path_split("a:b") == ["a", "b"]

    # If a single windows file path is passed on linux/osx it's not split on ":"
    assert utils.Platform.path_split(r"Z:\test", pathsep=":") == [r"Z:\test"]
    # TODO: This test covers the current behavior but ideally we can figure out
    # a way to prevent splitting the two windows file paths.
    assert utils.Platform.path_split(r"Z:\test:X:test", pathsep=":") == [
        "Z",
        r"\test",
        "X",
        "test",
    ]


@pytest.mark.parametrize("filename", ("merge_agnostic.json", "merge_specific.json"))
def test_merge(config_root, filename):
    """Test that we can merge multiple "*" dicts into a single dict."""
    json_path = config_root / "merge_dict" / filename

    with json_path.open() as fle:
        test_data = json.load(fle)

    # The data to feed the apply_platform_wildcards method
    file_one = test_data["file_one"]
    file_two = test_data["file_two"]
    # The output from the apply_platform_wildcards method to check against
    out_data = test_data["out_data"]

    merger = MergeDict()
    result = merger.apply_platform_wildcards(file_one)
    result = merger.apply_platform_wildcards(file_two, output=result)

    assert result == out_data



import datetime
import json
import logging
from pathlib import Path

from . import utils

logger = logging.getLogger(__name__)


class UriObj:
    def __init__(self, uri=None, timedout=False):
        self._uri = uri
        self._timedout = timedout

    def __str__(self):
        if self.uri is None:
            return ""
        else:
            return self.uri

    @property
    def timedout(self):
        return self._timedout

    @property
    def uri(self):
        return self._uri


class UserPrefs(dict):
    """Stores/restores hab user preferences in a json file."""

    def __init__(self, resolver):
        self.resolver = resolver
        self._is_loaded = False

    @property
    def enabled(self):
        """Returns if user preferences are enabled.

        The resolver's site config value "prefs_default" controls if preferences
        can be enabled and the default value(uses index 0). If set to "disabled"
        this will always return False. The cli enable/disable flags are used to
        enable or disable the default value(--prefs/--no-prefs). Defaults to
        "disabled".
        """
        try:
            return self._enabled
        except AttributeError:
            # This is the first call, return the default value
            self.enabled = None
            return self._enabled

    @enabled.setter
    def enabled(self, prefs):
        # Site resolves "prefs_default" into a list of options, the first item
        # in the list is respected
        default = self.resolver.site.get("prefs_default", [])
        if default:
            default = default[0]
            # Convert command flag strings to bolean values
            default = True if default == "--prefs" else default
            default = False if default == "--no-prefs" else default
        else:
            default = "disabled"

        if default == "disabled":
            # If the site disables preferences always use False
            logger.debug("Pref loading disabled by site")
            self._enabled = False
        elif prefs is None:
            logger.debug(f"Pref loading defaulting to {default} by site")
            self._enabled = default
        else:
            logger.debug(f"Pref loading set to {prefs}")
            self._enabled = prefs

    @property
    def filename(self):
        """The file path to load/save prefs from using json. Defaults to
        `Platform.user_prefs_filename()`.
        """
        try:
            return self._filename
        except AttributeError:
            self._filename = utils.Platform.user_prefs_filename()
            return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = Path(value)

    def load(self, force=False):
        """Load user preferences from self.filename if they haven't already
        been loaded. Does not use pyjson5 as this file is entirely managed by hab.

        Args:
            force (bool, optional): Force re-loading of preferences.

        Returns:
            bool: If self.filename exists.
        """
        if not self.filename.exists():
            return False

        if self._is_loaded and not force:
            return True

        with self.filename.open() as fle:
            try:
                # NOTE: This file is always saved using the native json library
                # there is no reason to support json5 encoded files here.
                data = json.load(fle)
            except ValueError as error:
                # When encountering corrupt saved user pref data, clear and
                # reset the user prefs. Warn about it this isn't entirely hidden
                logger.warning("User pref file corrupt, resetting.")
                logger.info("User pref exception suppressed:", exc_info=error)
                data = {}
            self.update(data)
            self._is_loaded = True
        return True

    def save(self):
        """Save user preferences to self.filename. Does not use pyjson5 as this
        file is entirely managed by hab.
        """
        with self.filename.open("w") as fle:
            return json.dump(self, fle, indent=4, cls=utils.HabJsonEncoder)

    @classmethod
    def _fromisoformat(cls, value):
        """Calls `datetime.fromisoforamt` if possible otherwise replicates
        its basic requirements (for python 3.6 support).
        """
        if isinstance(value, datetime.datetime):
            return value
        try:
            return datetime.datetime.fromisoformat(value)
        except AttributeError:
            iso_format = r"%Y-%m-%dT%H:%M:%S.%f"
            return datetime.datetime.strptime(value, iso_format)

    def uri_check(self):
        """Returns the uri saved in preferences. It will only do that if enabled
        and uri_is_timedout allow it. Returns None otherwise. This will call load
        to ensure the preference file has been loaded.
        """
        if self.enabled:
            # Ensure the preferences are loaded.
            self.load()
            uri = self.get("uri")
            if uri:
                return UriObj(uri, self.uri_is_timedout)
        return UriObj()

    @property
    def uri(self):
        return self.uri_check().uri

    @uri.setter
    def uri(self, uri):
        if self.enabled:
            # Ensure the preferences are loaded.
            self.load()
            self["uri"] = uri
            self["uri_last_changed"] = datetime.datetime.today()
            self.save()
            logger.debug(f"User prefs saved to {self.filename}")

    @property
    def uri_last_changed(self):
        value = self.get("uri_last_changed")
        if value is None:
            return value
        return self._fromisoformat(value)

    @property
    def uri_is_timedout(self):
        """Returns True if "uri_last_changed" is stored on prefs and that time is
        longer than uri_timeout."""

        uri_last_changed = self.uri_last_changed

        if uri_last_changed:
            current = datetime.datetime.today()
            timeout = self.uri_timeout
            if timeout and current - uri_last_changed > timeout:
                logger.debug(f"Saved URI exceeded timeout of {timeout}")
                return True

        return False

    @property
    def uri_timeout(self):
        """Returns a datetime.timedelta object if the site configuration defines
        the "prefs_uri_timeout" setting. If set, this should be a dictionary of
        kwargs to construct the timedelta. Otherwise returns None.
        """
        timeout = self.resolver.site.get("prefs_uri_timeout")
        if timeout:
            return datetime.timedelta(**timeout)
        return None



import json
from pathlib import Path

from hab import Resolver, Site
from hab.cache import Cache
from hab.parsers import Config, DistroVersion, HabBase


def test_cached_keys(uncached_resolver):
    check = {
        "config_paths": ("*.json", Config),
        "distro_paths": ("*/.hab.json", DistroVersion),
    }

    cache = uncached_resolver.site.cache
    # Container variable is not set by default
    assert not hasattr(cache, "_cached_keys")
    # On first call generates the correct return value
    assert cache.cached_keys == check
    # And stored the value in the container variable
    assert hasattr(cache, "_cached_keys")

    # Verify that the cached value is returned once generated
    cache._cached_keys = "Test value"
    assert cache.cached_keys == "Test value"


def test_site_cache_path(config_root, uncached_resolver, tmpdir):
    cache = uncached_resolver.site.cache

    # Test default
    site_file = Path(tmpdir) / "test.json"
    assert cache.cache_template == "{stem}.habcache"
    assert cache.site_cache_path(site_file) == tmpdir / "test.habcache"

    # Test that cache_template is respected
    cache.cache_template = "<{stem}>.ext"
    assert cache.site_cache_path(site_file) == tmpdir / "<test>.ext"

    # Test that `site_cache_file_template` site config setting is respected
    site = Site([config_root / "site" / "site_cache_file.json"])
    assert site.cache.cache_template == ".{stem}.hab_cache"


def test_save_cache(config_root, tmpdir, habcached_resolver):
    # Check that the habcache file generated the expected output text
    # Note: This file will need updated as the test configuration is updated
    check_path = config_root / "site_main_check.habcache"
    cache_file = habcached_resolver._test_cache_file
    check = check_path.open().readlines()
    cache = cache_file.open().readlines()
    # Add trailing white space to match template file's trailing white space
    cache[-1] += "\n"
    assert len(cache) == len(
        check
    ), f"Generated cache does not have the same number of lines: {cache_file}"

    for i in range(len(cache)):
        assert (
            cache[i] == check[i]
        ), f"Difference on line: {i} between the generated cache and {check_path}."


def test_load_cache(config_root, uncached_resolver, habcached_site_file):
    """Tests non-cached resolved data matches a reference cached version."""
    cached_site = Site([habcached_site_file])
    cached_resolver = Resolver(cached_site)

    # Load the reference cache. In normal operation _cache is None until the
    # first time `.cache` is called, but for this test we won't be doing that.
    cached_site.cache._cache = {}
    cached_site.cache.load_cache(config_root / "site_main_check.habcache")

    # Check that un-cached resolver settings match the reference cache
    for key in ("config_paths", "distro_paths"):
        assert getattr(uncached_resolver, key) == getattr(cached_resolver, key)


def test_unsupported_version_warning(uncached_resolver, tmpdir, caplog):
    cache_file = Path(tmpdir) / "test.habcache"
    warn_msg = (
        "File is using a unsupported habcache version {}. "
        f"Only versions > {{}} are supported, ignoring {cache_file}"
    )

    def _load_cache_version(version, cls=Cache):
        # Generate a test cache for the given version number
        with cache_file.open("w") as fle:
            json.dump({"version": version}, fle)

        cache = cls(uncached_resolver.site)
        cache._cache = {}
        caplog.clear()
        cache.load_cache(cache_file)
        return [rec.message for rec in caplog.records]

    # No warning if version is supported
    assert _load_cache_version(1) == []
    # Warning logged if cache version is newer than supported_version
    assert _load_cache_version(2) == [warn_msg.format(2, 1)]

    # Warning is logged correctly for higher supported versions
    class CacheV2(Cache):
        supported_version = 2

    assert _load_cache_version(1, cls=CacheV2) == []
    assert _load_cache_version(2, cls=CacheV2) == []
    assert _load_cache_version(3, cls=CacheV2) == [warn_msg.format(3, 2)]


def test_cached_method(config_root, habcached_site_file):
    """Test the Cache.cache method options."""
    cached_site = Site([habcached_site_file])
    cache = cached_site.cache
    assert cache.enabled is True

    # At this point _cache is None
    assert cache._cache is None
    # Simulate the cache already being loaded
    cache._cache = "Test value"
    assert cache.cache() == "Test value"

    # Forcing the cache to reload, re-generates a cache
    result = cache.cache(force=True)
    assert isinstance(result, dict)
    assert len(result)

    # Check that disabling caching causes the cache to not be returned
    cache.enabled = False
    assert cache.enabled is False
    result = cache.cache()
    assert isinstance(result, dict)
    assert not len(result)

    # Check that HabBase._cache returns a empty dict.
    assert HabBase(None, None)._cache() == {}


def test_resolver_cache(request, resolver):
    """Tests that cached and uncached resolvers actually use/don't use the cache.

    `uncached_resolver` should not have a habcache file and shouldn't use a cache.
    `habcached_resolver` should have a habcache and uses it.
    """
    # Figure out if this is the cached or uncached resolver test
    is_cached = "habcached_resolver" in request.fixturenames

    # Get the site file path
    assert len(resolver.site.paths) == 1
    site_file = resolver.site.paths[0]
    cache_file = resolver.site.cache.site_cache_path(site_file)

    # The .habcache file should only exist when testing cached
    if is_cached:
        assert cache_file.exists()
    else:
        assert not cache_file.exists()

    # force the resolver to load config/distro information
    resolver.resolve("not_set")

    cache = resolver.site.cache._cache
    if is_cached:
        # This habcache setup has exactly one cached glob string
        assert len(cache["config_paths"]) == 1
        assert len(cache["distro_paths"]) == 1
        # The flat cache has many configs/distros, the test only needs to ensure
        # that we have gotten some results
        assert len(cache["flat"]["config_paths"]) > 10
        assert len(cache["flat"]["distro_paths"]) > 10
    else:
        # If there aren't any habcache files, a default dict is returned
        assert cache == {"flat": {"config_paths": {}, "distro_paths": {}}}



import os
from pathlib import PurePosixPath, PureWindowsPath

import pytest

from hab import NotSet, Resolver, Site, utils
from hab.parsers import UnfrozenConfig


def update_config(check, config_root, platform):
    """Correctly fill in {config_root} for the given frozen config.
    This allows the test to pass no matter what os it is run on.

    Args:
        check (dict): A frozen config loaded from a test json file.
        config_root (pathlib.Path): The config_root string format variable is
            replaced with this value if it matches platform. Otherwise the generic
            "c:\\" for windows and "/hab" for linux is used.
        platform (str): The current platform the test is running on.
    """
    env = check["environment"]
    for plat in env:
        if plat == platform:
            cfg_root = config_root
        else:
            cfg_root = "c:" if plat == "windows" else "/hab"
        for k, values in env[plat].items():
            for i, v in enumerate(values):
                env[plat][k][i] = v.format(config_root=cfg_root)


def test_json_dumps():
    """Check that dumps_json returns the expected results for non-standard
    objects."""
    data = {"NotSet": NotSet}
    assert utils.dumps_json(data) == '{"NotSet": null}'
    assert utils.dumps_json(data, indent=2) == "\n".join(
        [
            "{",
            '  "NotSet": null',
            "}",
        ]
    )


@pytest.mark.parametrize("platform,pathsep", (("win32", ";"), ("linux", ":")))
def test_freeze(monkeypatch, config_root, platform, pathsep):
    monkeypatch.setattr(utils, "Platform", utils.WinPlatform)
    monkeypatch.setattr(os, "pathsep", pathsep)
    site = Site([config_root / "site_main.json"])
    resolver = Resolver(site=site)
    cfg_root = utils.path_forward_slash(config_root)

    # Add a platform_path_maps mapping to convert the current hab checkout path
    # to a generic know path on the other platform for uniform testing.
    mappings = site.frozen_data[site.platform]["platform_path_maps"]
    mappings["local-hab"] = {
        "linux": PurePosixPath("/hab"),
        "windows": PureWindowsPath("c:/"),
    }
    # Preserve the current platform's path so it matches the frozen output
    if site.platform == "windows":
        mappings["local-hab"]["windows"] = PureWindowsPath(cfg_root)
    else:
        mappings["local-hab"][site.platform] = PurePosixPath(cfg_root)

    # Resolve the URI for frozen testing
    cfg = resolver.resolve("not_set/distros")

    # Ensure consistent testing across platforms. cfg has the current os's
    # file paths instead of what is stored in frozen.json
    cfg.frozen_data["aliases"]["linux"]["dcc"] = "TEST_DIR_NAME//the_dcc"
    cfg.frozen_data["aliases"]["windows"]["dcc"] = "TEST_DIR_NAME\\the_dcc.exe"

    # Ensure the HAB_URI environment variable is defined on the FlatConfig object
    # When checking the return from `cfg.freeze()` below HAB_URI is removed to
    # simplify the output json data.
    assert cfg.frozen_data["environment"]["linux"]["HAB_URI"] == ["not_set/distros"]
    assert cfg.frozen_data["environment"]["windows"]["HAB_URI"] == ["not_set/distros"]

    ret = cfg.freeze()
    check_file = config_root / "frozen.json"
    check = utils.json.load(check_file.open())
    # Apply template values so we can easily check against frozen.
    update_config(check, cfg_root, site.platform)

    assert ret == check

    # Check that optional properties are excluded from the dictionary if empty
    ret = cfg.freeze()
    assert "versions" in ret
    # Simulate having versions being empty
    cfg.frozen_data["versions"] = []
    ret = cfg.freeze()
    assert "versions" not in ret

    # Frozen configs don't need to encode alias_mods, the modifications are
    # already baked into aliases
    assert "alias_mods" not in ret


def test_unfreeze(config_root, resolver):
    check_file = config_root / "frozen.json"

    # Note: For this test, we don't need to worry about "{config_root}" templates.
    frozen_config = utils.json.load(check_file.open())
    cfg = UnfrozenConfig(frozen_config, resolver)

    assert cfg.context == frozen_config["context"]
    assert cfg.name == frozen_config["name"]
    assert cfg.uri == frozen_config["uri"]
    assert (
        cfg.frozen_data["aliases"]["linux"]["dcc"]
        == frozen_config["aliases"]["linux"]["dcc"]
    )
    assert (
        cfg.frozen_data["aliases"]["windows"]["dcc"]
        == frozen_config["aliases"]["windows"]["dcc"]
    )

    assert cfg.versions == frozen_config["versions"]

    # HAB_URI is removed from the frozen data. Storing it per platform would
    # just make the frozen string longer. Re-add these to the check data so it
    # matches UnfrozenConfig also adding it.
    frozen_config["environment"]["linux"]["HAB_URI"] = "not_set/distros"
    frozen_config["environment"]["windows"]["HAB_URI"] = "not_set/distros"
    # Check that environment was restored correctly
    assert cfg.frozen_data["environment"] == frozen_config["environment"]

    # Check various class overrides
    assert cfg._dump_versions(cfg.versions) == sorted(frozen_config["versions"])
    assert cfg.aliases == frozen_config["aliases"][utils.Platform.name()]
    assert "dcc" in cfg.aliases
    assert cfg.fullpath == "not_set/distros"
    assert cfg.inherits is False

    # Frozen configs don't need to encode alias_mods, the modifications are
    # already baked into aliases
    assert cfg.alias_mods is NotSet


def test_decode_freeze(config_root):
    check_file = config_root / "frozen_no_distros.json"
    checks = utils.json.load(check_file.open())
    v1 = checks["version1"]
    raw = checks["raw"]

    # Check that supported freeze's are decoded correctly
    assert utils.decode_freeze(v1) == raw
    assert utils.decode_freeze(checks["version2"]) == raw

    # Check that padded versions are also supported
    assert utils.decode_freeze(f"v01:{v1[3:]}") == raw

    # Check that non-versioned freeze strings raise an helpful exception
    for check in (
        # Missing `v1:`
        v1[3:],
        # Missing `v'
        f"1:{v1[3:]}",
    ):
        with pytest.raises(
            ValueError, match=r"Missing freeze version information in format `v0:...`"
        ):
            utils.decode_freeze(check)

    # Check that versions other than numbers raise a helpful exception
    with pytest.raises(ValueError, match=r"Version INVALID is not valid."):
        utils.decode_freeze(f"vINVALID:{v1[3:]}")

    # check that other version encodings return nothing
    assert utils.decode_freeze(f"v3:{v1[3:]}") is None
    assert utils.decode_freeze(f"v0:{v1[3:]}") is None


def test_encode_freeze(config_root, resolver):
    cfg = resolver.resolve("not_set/no_distros")
    check_file = config_root / "frozen_no_distros.json"
    checks = utils.json.load(check_file.open())

    # Check that the dict contains the expected contents
    freeze = cfg.freeze()
    assert freeze == checks["raw"]

    # Check that supported version encodings are correct
    version1 = utils.encode_freeze(freeze, version=1)
    assert version1 == checks["version1"]
    version2 = utils.encode_freeze(freeze, version=2)
    assert version2 == checks["version2"]

    # Check that if a version is not defined, the default(2) is used.
    # This is used if `Site.get("freeze_version")` is not defined.
    version_none = utils.encode_freeze(freeze, version=None)
    assert version_none == checks["version2"]

    # Check that other version encodings return nothing
    assert utils.encode_freeze(freeze, version=0) is None
    assert utils.encode_freeze(freeze, version=3) is None

    # Check that site kwarg is respected
    site = Site()
    # Force the site to version 1, but version is not specified
    site["freeze_version"] = 1
    version1 = utils.encode_freeze(freeze, version=None, site=site)
    assert version1 == checks["version1"]

    # If version is passed, site is ignored
    version1 = utils.encode_freeze(freeze, version=2, site=site)
    assert version1 == checks["version2"]
